# encoding: utf-8

import numpy as np
import pytest

from cr.cube.crunch_cube import CrunchCube

from ...fixtures import CR, SM  # ---mnemonic: SM = 'scale means'---

from . import assert_scale_means_equal


def test_ca_cat_x_items():
    cube = CrunchCube(SM.CA_CAT_X_ITEMS)
    expected = [[np.array([1.50454821, 3.11233766, 3.35788192, 3.33271833]), None]]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    slice_ = cube.slices[0]
    actual = slice_.scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test ScaleMeans marginal
    actual = slice_.scale_means_margin(0)
    assert actual is None
    with pytest.raises(ValueError):
        # Items dimension doesn't have means
        actual = slice_.scale_means_margin(1)


def test_ca_items_x_cat():
    cube = CrunchCube(SM.CA_ITEMS_X_CAT)
    expected = [[None, np.array([1.50454821, 3.11233766, 3.35788192, 3.33271833])]]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    slice_ = cube.slices[0]
    actual = slice_.scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test ScaleMeans marginal
    with pytest.raises(ValueError):
        # Items dimension doesn't have means
        actual = slice_.scale_means_margin(0)
    actual = slice_.scale_means_margin(1)
    assert actual is None


def test_ca_x_mr():
    cube = CrunchCube(SM.CA_X_MR)
    expected = [
        [np.array([1.29787234, 1.8, 1.48730964, np.nan]), None],
        [np.array([3.31746032, 3.10743802, 3.09976976, np.nan]), None],
        [np.array([3.31205674, 3.23913043, 3.37745455, np.nan]), None],
        [np.array([3.53676471, 3.34814815, 3.3147877, np.nan]), None],
    ]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    slice_ = cube.slices[0]
    actual = slice_.scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test ScaleMeans marginal
    actual = slice_.scale_means_margin(axis=0)
    assert actual is None
    actual = slice_.scale_means_margin(axis=1)
    assert actual == 1.504548211036992


def test_cat_x_ca_cat_x_items():
    cube = CrunchCube(SM.CAT_X_CA_CAT_X_ITEMS)
    expected = [
        [np.array([1.34545455, 2.46938776, 2.7037037, 2.65454545]), None],
        [np.array([1.41935484, 3.25663717, 3.48, 3.58536585]), None],
        [np.array([1.49429038, 3.44905009, 3.59344262, 3.53630363]), None],
        [np.array([1.43365696, 3.02816901, 3.37987013, 3.32107023]), None],
        [np.array([1.22670025, 2.49473684, 2.79848866, 2.78987342]), None],
        [np.array([2.53061224, 3.68421053, 3.9862069, 4.03472222]), None],
    ]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    actual = cube.scale_means(hs_dims=[0, 1])
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    slice_ = cube.slices[0]
    actual = slice_.scale_means()
    assert_scale_means_equal(actual, expected[0])
    actual = slice_.scale_means(hs_dims=[0, 1])
    assert_scale_means_equal(actual, expected[0])

    # Test ScaleMeans marginal
    actual = slice_.scale_means_margin(axis=0)
    assert actual is None
    with pytest.raises(ValueError):
        slice_.scale_means_margin(axis=1)


def test_cat_x_cat():
    cube = CrunchCube(SM.CAT_X_CAT)
    expected = [
        [
            np.array([2.6009281, 2.3522267, 2.3197279, 3.3949192]),
            np.array(
                [1.43636364, 2.45238095, 2.4730832, 2.68387097, 2.8375, 2.15540541]
            ),
        ]
    ]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    slice_ = cube.slices[0]
    actual = slice_.scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test ScaleMeans marginal
    actual = slice_.scale_means_margin(0)
    assert actual == 2.536319612590799
    actual = slice_.scale_means_margin(1)
    assert actual == 2.6846246973365617


def test_cat_x_mr():
    cube = CrunchCube(SM.CAT_X_MR)
    expected = [[np.array([2.45070423, 2.54471545, 2.54263006, np.nan]), None]]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    slice_ = cube.slices[0]
    actual = slice_.scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test ScaleMeans marginal
    actual = slice_.scale_means_margin(0)
    assert actual is None
    actual = slice_.scale_means_margin(1)
    assert actual == 2.5323565323565322


def test_mr_x_cat():
    cube = CrunchCube(SM.MR_X_CAT)
    expected = [[None, np.array([2.45070423, 2.54471545, 2.54263006, np.nan])]]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    slice_ = cube.slices[0]
    actual = slice_.scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test ScaleMeans marginal
    actual = slice_.scale_means_margin(0)
    assert actual == 2.5323565323565322
    actual = slice_.scale_means_margin(1)
    assert actual is None


def test_univariate_cat():
    cube = CrunchCube(SM.UNIVARIATE_CAT)
    expected = [[np.array([2.6865854])]]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    slice_ = cube.slices[0]
    actual = slice_.scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test ScaleMeans marginal
    with pytest.raises(ValueError):
        actual = slice_.scale_means_margin(0)


def test_cat_x_cat_with_hs():
    cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)

    # Test without H&S
    expected = [
        [
            np.array(
                [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85]
            ),
            np.array([3.87368421, 2.51767677, 3.38429752, 3.66666667, 4.13235294]),
        ]
    ]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    slice_ = cube.slices[0]
    actual = slice_.scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test with H&S
    expected = [
        [
            np.array(
                [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85]
            ),
            np.array(
                [3.87368421, 2.51767677, np.nan, 3.38429752, 3.66666667, 4.13235294]
            ),
        ]
    ]
    actual = cube.scale_means(hs_dims=[0, 1])
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    actual = cube.slices[0].scale_means(hs_dims=[0, 1])
    assert_scale_means_equal(actual, expected[0])


def test_univariate_with_hs():
    cube = CrunchCube(CR.ECON_BLAME_WITH_HS)

    # Test without H&S
    expected = [[np.array([2.17352056])]]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    actual = cube.slices[0].scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test with H&S
    expected = [[np.array([2.17352056])]]
    actual = cube.scale_means(hs_dims=[0])
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    actual = cube.slices[0].scale_means(hs_dims=[0, 1])
    assert_scale_means_equal(actual, expected[0])


def test_cat_x_cat_with_hs_on_both_dims():
    cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS)

    # Test without H&S
    expected = [
        [
            np.array(
                [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85]
            ),
            np.array([3.87368421, 2.51767677, 3.38429752, 3.66666667, 4.13235294]),
        ]
    ]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    actual = cube.slices[0].scale_means()
    assert_scale_means_equal(actual, expected[0])

    # Test with H&S
    expected = [
        [
            np.array(
                [
                    2.19444444,
                    2.19230769,
                    2.26666667,
                    np.nan,
                    1.88990826,
                    1.76363636,
                    3.85,
                ]
            ),
            np.array(
                [3.87368421, 2.51767677, np.nan, 3.38429752, 3.66666667, 4.13235294]
            ),
        ]
    ]
    actual = cube.scale_means(hs_dims=[0, 1])
    assert_scale_means_equal(actual, expected)

    # Test for cube slices
    actual = cube.slices[0].scale_means(hs_dims=[0, 1])
    assert_scale_means_equal(actual, expected[0])


def test_ca_x_mr_with_hs_and_pruning():
    cube = CrunchCube(CR.CA_X_MR_HS)
    expected = [
        [np.array([2.50818336, 2.56844883, 2.90251939, np.nan]), None],
        [np.array([2.78385708, 2.69292009, 3.11594714, np.nan]), None],
        [np.array([np.nan, np.nan, np.nan, np.nan]), None],
    ]
    actual = cube.scale_means()
    assert_scale_means_equal(actual, expected)
    expected = [
        [np.array([2.50818336, 2.56844883, 2.90251939]), None],
        [np.array([2.78385708, 2.69292009, 3.11594714]), None],
        [np.array([]), None],
    ]
    actual = cube.scale_means(prune=True)
    assert_scale_means_equal(actual, expected)
    actual = cube.scale_means(prune=True, hs_dims=[0, 1])
    assert_scale_means_equal(actual, expected)


def test_cat_x_cat_pruning_and_hs():
    cube = CrunchCube(CR.CAT_X_CAT_PRUNING_HS)
    expected = [
        [
            np.array([1.57933884, 2.10618401, 2.30460074, np.nan, 2.34680135]),
            np.array([1.74213625, 1.97, 2.45356177, 2.11838791, np.nan, 2.0]),
        ]
    ]
    actual = cube.scale_means()

    # Just H&S
    assert_scale_means_equal(actual, expected)
    expected = [
        [
            np.array([1.57933884, np.nan, 2.10618401, 2.30460074, np.nan, 2.34680135]),
            np.array([1.74213625, np.nan, 1.97, 2.45356177, 2.11838791, np.nan, 2.0]),
        ]
    ]
    actual = cube.scale_means(hs_dims=[0, 1])

    # Just pruning
    assert_scale_means_equal(actual, expected)
    expected = [
        [
            np.array([1.57933884, 2.10618401, 2.30460074, 2.34680135]),
            np.array([1.74213625, 1.97, 2.45356177, 2.11838791, 2.0]),
        ]
    ]
    actual = cube.scale_means(prune=True)
    assert_scale_means_equal(actual, expected)

    # Pruning and H&S
    assert_scale_means_equal(actual, expected)
    expected = [
        [
            np.array([1.57933884, np.nan, 2.10618401, 2.30460074, 2.34680135]),
            np.array([1.74213625, np.nan, 1.97, 2.45356177, 2.11838791, 2.0]),
        ]
    ]
    actual = cube.scale_means(hs_dims=[0, 1], prune=True)
    assert_scale_means_equal(actual, expected)


def test_cat_x_cat_scale_means_margin():
    cs = CrunchCube(SM.CAT_X_CAT_SM_MARGIN).slices[0]
    expected = 2.6846246973365617
    assert cs.scale_means_margin(1) == expected

    expected = 2.536319612590799
    assert cs.scale_means_margin(0) == expected


def test_cat_single_element_x_cat():
    cs = CrunchCube(SM.CAT_SINGLE_ELEMENT_X_CAT).slices[0]
    scale_means = cs.scale_means()
    np.testing.assert_equal(scale_means[0], np.array([np.nan, np.nan, np.nan, np.nan]))
    np.testing.assert_equal(scale_means[1], np.array([np.nan]))



# encoding: utf-8

"""Provides the Index function object class."""

from __future__ import division

import numpy as np


class Index(object):
    """Function object providing index calculations for a CrunchCube object.

    It's only interface is its `.data()` classmethod. It is not intended to
    be instantiated directly.
    """

    @classmethod
    def data(cls, cube, weighted, prune):
        """Return ndarray representing table index by margin."""
        return cls()._data(cube, weighted, prune)

    def _data(self, cube, weighted, prune):
        """ndarray representing table index by margin."""
        result = []
        for slice_ in cube.slices:
            if cube.has_mr:
                return self._mr_index(cube, weighted, prune)
            num = slice_.margin(axis=0, weighted=weighted, prune=prune)
            den = slice_.margin(weighted=weighted, prune=prune)
            margin = num / den
            proportions = slice_.proportions(axis=1, weighted=weighted, prune=prune)
            result.append(proportions / margin)

        if len(result) == 1 and cube.ndim < 3:
            result = result[0]
        else:
            if prune:
                mask = np.array([slice_.mask for slice_ in result])
                result = np.ma.masked_array(result, mask)
            else:
                result = np.array(result)

        return result

    def _mr_index(self, cube, weighted, prune):
        # mr by mr
        if len(cube.dimensions) == 2 and cube.mr_dim_ind == (0, 1):
            col_proportions = cube.proportions(axis=0, weighted=weighted, prune=prune)
            row_proportions = cube.proportions(axis=1, weighted=weighted, prune=prune)
            return col_proportions / row_proportions

        # mr by cat and cat by mr
        if cube.mr_dim_ind == 0 or cube.mr_dim_ind == 1:
            axis = cube.mr_dim_ind
            num = cube.margin(axis=1 - axis, weighted=weighted, prune=prune)
            den = cube.margin(weighted=weighted, prune=prune)
            margin = num / den
            proportions = cube.proportions(axis=axis, weighted=weighted, prune=prune)
            if cube.mr_dim_ind == 0:
                margin = margin[:, np.newaxis]  # pivot
            return proportions / margin

        raise ValueError("Unexpected dimension types for cube with MR.")



# encoding: utf-8

"""Unit test suite for the  cr.cube.measures.scale_means module."""

from cr.cube.scalar import MeansScalar


class Describe_MeansScalar(object):
    def it_knows_its_ndim(self):
        assert MeansScalar(None, None).ndim == 0



# encoding: utf-8

"""Provides the CubeSet and Cube classes.

CubeSet is the main API class for manipulating Crunch.io JSON cube responses.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import json

import numpy as np

from cr.cube.cubepart import CubePartition
from cr.cube.dimension import AllDimensions
from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.util import lazyproperty

np.seterr(divide="ignore", invalid="ignore")


class CubeSet(object):
    """Represents a multi-cube cube-response.

    Also works just fine for a single cube-response passed inside a sequence, allowing
    uniform handling of single and multi-cube responses.

    `cube_responses` is a sequence of cube-response dicts received from Crunch. The
    sequence can contain a single item, such as a cube-response for a slide, but it must
    be contained in a sequence. A tabbook cube-response sequence can be passed as it was
    received.

    `transforms` is a sequence of transforms dicts corresponding in order to the
    cube-responses. `population` is the estimated target population and is used when
    a population-projection measure is requested. `min_base` is an integer representing
    the minimum sample-size used for indicating values that are unreliable by reason of
    insufficient sample (base).
    """

    def __init__(self, cube_responses, transforms, population, min_base):
        self._cube_responses = cube_responses
        self._transforms_dicts = transforms
        self._population = population
        self._min_base = min_base

    @lazyproperty
    def can_show_pairwise(self):
        """True if all 2D cubes in a multi-cube set can provide pairwise comparison."""
        if len(self._cubes) < 2:
            return False

        return all(
            all(dt in DT.ALLOWED_PAIRWISE_TYPES for dt in cube.dimension_types[-2:])
            and cube.ndim >= 2
            for cube in self._cubes[1:]
        )

    @lazyproperty
    def description(self):
        """str description of first cube in this set."""
        return self._cubes[0].description

    @lazyproperty
    def has_means(self):
        """True if cubes in this set include a means measure."""
        return self._cubes[0].has_means

    @lazyproperty
    def has_weighted_counts(self):
        """True if cube-responses include a weighted-count measure."""
        return self._cubes[0].is_weighted

    @lazyproperty
    def is_ca_as_0th(self):
        """True for multi-cube when first cube represents a categorical-array.

        A "CA-as-0th" tabbook tab is "3D" in the sense it is "sliced" into one table
        (partition-set) for each of the CA subvariables.
        """
        # ---can only be true for multi-cube case---
        if not self._is_multi_cube:
            return False
        # ---the rest depends on the row-var cube---
        cube = self._cubes[0]
        # ---True if row-var cube is CA---
        return cube.dimension_types[0] == DT.CA_SUBVAR

    @lazyproperty
    def missing_count(self):
        """The number of missing values from first cube in this set."""
        return self._cubes[0].missing

    @lazyproperty
    def name(self):
        """str name of first cube in this set."""
        return self._cubes[0].name

    @lazyproperty
    def partition_sets(self):
        """Sequence of cube-partition collections across all cubes of this cube-set.

        This value might look like the following for a ca-as-0th tabbook, for example:

            (
                (_Strand, _Slice, _Slice),
                (_Strand, _Slice, _Slice),
                (_Strand, _Slice, _Slice),
            )

        and might often look like this for a typical slide:

            ((_Slice,))

        Each partition set represents the partitions for a single "stacked" table. A 2D
        slide has a single partition-set of a single _Slice object, as in the second
        example above. A 3D slide would have multiple partition sets, each of a single
        _Slice. A tabook will have multiple partitions in each set, the first being
        a _Strand and the rest being _Slice objects. Multiple partition sets only arise
        for a tabbook in the CA-as-0th case.
        """
        return tuple(zip(*(cube.partitions for cube in self._cubes)))

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for this cube-set.

        This value is required for properly calculating population on a cube where
        a filter has been applied. Returns 1.0 for an unfiltered cube. Returns `np.nan`
        if the unfiltered count is zero, which would otherwise result in
        a divide-by-zero error.
        """
        return self._cubes[0].population_fraction

    @lazyproperty
    def _cubes(self):
        """Sequence of Cube objects containing data for this analysis."""
        return tuple(self._iter_cubes())

    @lazyproperty
    def _is_multi_cube(self):
        """True if more than one cube-response was provided on construction."""
        return len(self._cube_responses) > 1

    @lazyproperty
    def _is_numeric_mean(self):
        """True when CubeSet is special-case "numeric-mean" case requiring inflation.

        When a numeric variable appears as the rows-dimension in a multitable analysis,
        its cube-result has been "reduced" to the mean-value of those numerics. This is
        in contrast to being "bucketized" into an arbitrary set of numeric-range
        categories like 0-5, 5-10, etc. In the process, as an artifact of the ZZ9 query
        response, that dimension is removed. As a result, the rows-dimension cube is 0D
        and the column-dimension cubes are 1D. These need to be "inflated" to restore
        the lost dimension such that they are uniform with other cube-results and can be
        processed without special-case code.

        "Inflation" is basically prefixing "1 x" to the dimensionality, for example a 1D
        of size 5 becomes a 1 x 5 2D result. Note this requires no mapping in the actual
        values because 5 = 1 x 5 = 5 (values).
        """
        # --- this case only arises in a multitable analysis ---
        if not self._is_multi_cube:
            return False

        # --- We need the cube to tell us the dimensionality. This redundant
        # --- construction is low-overhead because all Cube properties are lazy.
        return Cube(self._cube_responses[0]).ndim == 0

    def _iter_cubes(self):
        """Generate a Cube object for each of cube_responses.

        0D cube-responses and 1D second-and-later cubes are "inflated" to add their
        missing row dimension.
        """
        for idx, cube_response in enumerate(self._cube_responses):
            cube = Cube(
                cube_response,
                cube_idx=idx if self._is_multi_cube else None,
                transforms=self._transforms_dicts[idx],
                population=self._population,
                mask_size=self._min_base,
            )
            # --- all numeric-mean cubes require inflation to restore their
            # --- rows-dimension, others don't
            yield cube.inflate() if self._is_numeric_mean else cube


class Cube(object):
    """Provides access to individual slices on a cube-result.

    It also provides some attributes of the overall cube-result.

    `cube_idx` must be `None` (or omitted) for a single-cube CubeSet. This indicates the
    CubeSet contains only a single cube and influences behaviors like CA-as-0th.
    """

    def __init__(
        self, response, cube_idx=None, transforms=None, population=None, mask_size=0
    ):
        self._cube_response_arg = response
        self._transforms_dict = {} if transforms is None else transforms
        self._cube_idx_arg = cube_idx
        self._population = 0 if population is None else population
        self._mask_size = mask_size

    def __repr__(self):
        """Provide text representation suitable for working at console.

        Falls back to a default repr on exception, such as might occur in
        unit tests where object need not otherwise be provided with all
        instance variable values.
        """
        try:
            dimensionality = " x ".join(dt.name for dt in self.dimension_types)
            return "%s(name='%s', dimension_types='%s')" % (
                type(self).__name__,
                self.name,
                dimensionality,
            )
        except Exception:
            return super(Cube, self).__repr__()

    @lazyproperty
    def counts(self):
        return self.counts_with_missings[self._valid_idxs]

    @lazyproperty
    def counts_with_missings(self):
        return self._measure(self.is_weighted).raw_cube_array

    @lazyproperty
    def cube_index(self):
        """Offset of this cube within its CubeSet."""
        return 0 if self._cube_idx_arg is None else self._cube_idx_arg

    @lazyproperty
    def description(self):
        """Return the description of the cube."""
        if not self.dimensions:
            return None
        return self.dimensions[0].description

    @lazyproperty
    def dimension_types(self):
        """Tuple of DIMENSION_TYPE member for each dimension of cube."""
        return tuple(d.dimension_type for d in self.dimensions)

    @lazyproperty
    def dimensions(self):
        """_ApparentDimensions object providing access to visible dimensions.

        A cube involving a multiple-response (MR) variable has two dimensions
        for that variable (subvariables and categories dimensions), but is
        "collapsed" into a single effective dimension for cube-user purposes
        (its categories dimension is supressed). This collection will contain
        a single dimension for each MR variable and therefore may have fewer
        dimensions than appear in the cube response.
        """
        return self._all_dimensions.apparent_dimensions

    @lazyproperty
    def has_means(self):
        """True if cube includes a means measure."""
        return self._measures.means is not None

    def inflate(self):
        """Return new Cube object with rows-dimension added.

        A multi-cube (tabbook) response formed from a function (e.g. mean()) on
        a numeric variable arrives without a rows-dimension.
        """
        cube_dict = self._cube_dict
        dimensions = cube_dict["result"]["dimensions"]
        rows_dimension = {
            "references": {"alias": "mean", "name": "mean"},
            "type": {"categories": [{"id": 1, "name": "Mean"}], "class": "categorical"},
        }
        dimensions.insert(0, rows_dimension)
        return Cube(
            cube_dict,
            self._cube_idx_arg,
            self._transforms_dict,
            self._population,
            self._mask_size,
        )

    @lazyproperty
    def is_mr_aug(self):
        """True if cube is MR_AUG (augmented).

        An augmented-MR cube is required to compute pairwise-t-tests when an MR
        dimension is involved because that introduces *overlap* which otherwise inflates
        the case count.

        An MR_AUG cube is used to calculate statistical significance of complete-cases,
        and is created by artificially *augmenting* the original cube's dimensions, such
        as MR, or CAT_X_MR, by repeating the MR dimension to make the cube MR_X_MR or
        CAT_X_MR_X_MR for these respective examples. The repeating of the MR dimension
        allows the overlaps to be computed.
        """
        return (
            # --- there are at least three dimensions ---
            self.ndim >= 3
            # --- the last two are both MR ---
            and all(dim_type == DT.MR for dim_type in self.dimension_types[-2:])
            # --- and they both have the same alias ---
            and len(set(dim.alias for dim in self.dimensions[-2:])) == 1
        )

    @lazyproperty
    def is_weighted(self):
        """True if cube response contains weighted data."""
        return self._measures.is_weighted

    @lazyproperty
    def missing(self):
        """Get missing count of a cube."""
        return self._measures.missing_count

    @lazyproperty
    def name(self):
        """Return the name of the cube.

        If the cube has 2 diensions, return the name of the second one. In case
        of a different number of dimensions, default to returning the name of
        the last one. In case of no dimensions, return the empty string.
        """
        if not self.dimensions:
            return None
        return self.dimensions[0].name

    @lazyproperty
    def ndim(self):
        """int count of dimensions for this cube."""
        return len(self.dimensions)

    @lazyproperty
    def partitions(self):
        """Sequence of _Slice, _Strand, or _Nub objects from this cube-result."""
        return tuple(
            CubePartition.factory(
                self,
                slice_idx=slice_idx,
                transforms=self._transforms_dict,
                population=self._population,
                ca_as_0th=self._ca_as_0th,
                mask_size=self._mask_size,
            )
            for slice_idx in self._slice_idxs
        )

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for cube response.

        This value is required for properly calculating population on a cube
        where a filter has been applied. Returns 1.0 for an unfiltered cube.
        Returns `np.nan` if the unfiltered count is zero, which would
        otherwise result in a divide-by-zero error.
        """
        return self._measures.population_fraction

    @lazyproperty
    def title(self):
        """str alternate-name given to cube-result.

        This value is suitable for naming a Strand when displayed as a column. In this
        use-case it is a stand-in for the columns-dimension name since a strand has no
        columns dimension.
        """
        return self._cube_dict["result"].get("title", "Untitled")

    @lazyproperty
    def unweighted_counts(self):
        """ndarray of unweighted counts, valid elements only.

        Unweighted counts are drawn from the `result.counts` field of the cube result.
        These counts are always present, even when the measure is `mean` and there are
        no count measures. These counts are always unweighted, regardless of whether the
        cube is "weighted".
        """
        return self._measures.unweighted_counts.raw_cube_array[self._valid_idxs]

    @lazyproperty
    def _all_dimensions(self):
        """The AllDimensions object for this cube.

        The AllDimensions object provides access to all the dimensions appearing in the
        cube response, not only apparent dimensions (those that appear to a user). It
        also provides access to an _ApparentDimensions object which contains only those
        user-apparent dimensions (basically the categories dimension of each MR
        dimension-pair is suppressed).
        """
        return AllDimensions(dimension_dicts=self._cube_dict["result"]["dimensions"])

    @lazyproperty
    def _ca_as_0th(self):
        """True if slicing is to be performed in so-called "CA-as-0th" mode.

        In this mode, a categorical-array (CA) cube (2D) is sliced into a sequence of 1D
        slices, each of which represents one subvariable of the CA variable. Normally,
        a 2D cube-result becomes a single slice.
        """
        return (
            (self._cube_idx_arg == 0 or self._is_single_filter_col_cube)
            and len(self.dimension_types) > 0
            and self.dimension_types[0] == DT.CA
        )

    @lazyproperty
    def _cube_dict(self):
        """dict containing raw cube response, parsed from JSON payload."""
        try:
            cube_response = self._cube_response_arg
            # ---parse JSON to a dict when constructed with JSON---
            cube_dict = (
                cube_response
                if isinstance(cube_response, dict)
                else json.loads(cube_response)
            )
            # ---cube is 'value' item in a shoji response---
            return cube_dict.get("value", cube_dict)
        except TypeError:
            raise TypeError(
                "Unsupported type <%s> provided. Cube response must be JSON "
                "(str) or dict." % type(self._cube_response_arg).__name__
            )

    @lazyproperty
    def _is_single_filter_col_cube(self):
        """bool determines if it is a single column filter cube."""
        return self._cube_dict["result"].get("is_single_col_cube", False)

    def _measure(self, weighted):
        """_BaseMeasure subclass representing primary measure for this cube.

        If the cube response includes a means measure, the return value is
        means. Otherwise it is counts, with the choice between weighted or
        unweighted determined by *weighted*.

        Note that weighted counts are provided on an "as-available" basis.
        When *weighted* is True and the cube response is not weighted,
        unweighted counts are returned.
        """
        return (
            self._measures.means
            if self._measures.means is not None
            else self._measures.weighted_counts
            if weighted
            else self._measures.unweighted_counts
        )

    @lazyproperty
    def _measures(self):
        """_Measures object for this cube.

        Provides access to unweighted counts, and weighted counts and/or means
        when available.
        """
        return _Measures(self._cube_dict, self._all_dimensions)

    @lazyproperty
    def _slice_idxs(self):
        """Iterable of contiguous int indicies for slices to be produced.

        This value is to help cube-section construction which does not by itself know
        how many slices are in a cube-result.
        """
        if self.ndim < 3 and not self._ca_as_0th:
            return (0,)
        return range(len(self.dimensions[0].valid_elements))

    @lazyproperty
    def _valid_idxs(self):
        return np.ix_(
            *tuple(d.valid_elements.element_idxs for d in self._all_dimensions)
        )


class _Measures(object):
    """Provides access to measures contained in cube response."""

    def __init__(self, cube_dict, all_dimensions):
        self._cube_dict = cube_dict
        self._all_dimensions = all_dimensions

    @lazyproperty
    def is_weighted(self):
        """True if weights have been applied to the measure(s) for this cube.

        Unweighted counts are available for all cubes. Weighting applies to
        any other measures provided by the cube.
        """
        cube_dict = self._cube_dict
        if cube_dict.get("query", {}).get("weight") is not None:
            return True
        if cube_dict.get("weight_var") is not None:
            return True
        if cube_dict.get("weight_url") is not None:
            return True
        unweighted_counts = cube_dict["result"]["counts"]
        count_data = cube_dict["result"]["measures"].get("count", {}).get("data")
        if unweighted_counts != count_data:
            return True
        return False

    @lazyproperty
    def means(self):
        """_MeanMeasure object providing access to means values.

        None when the cube response does not contain a mean measure.
        """
        mean_measure_dict = (
            self._cube_dict.get("result", {}).get("measures", {}).get("mean")
        )
        if mean_measure_dict is None:
            return None
        return _MeanMeasure(self._cube_dict, self._all_dimensions)

    @lazyproperty
    def missing_count(self):
        """numeric representing count of missing rows in cube response."""
        if self.means:
            return self.means.missing_count
        return self._cube_dict["result"].get("missing", 0)

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for cube response.

        The filtered counts are calculated for complete-cases. This means that only the
        non-missing entries are included in the filtered counts. Complete cases are
        used only if the corresponding cases are included in the cube response. If not,
        the old-style default calculation is used.

        This value is required for properly calculating population on a cube
        where a filter has been applied. Returns 1.0 for an unfiltered cube.
        Returns `np.nan` if the unfiltered count is zero, which would
        otherwise result in a divide-by-zero error.
        """

        # Try and get the new-style complete-cases filtered counts
        filter_stats = (
            self._cube_dict["result"]
            .get("filter_stats", {})
            .get("filtered_complete", {})
            .get("weighted")
        )
        if filter_stats:
            # If new format is present in response json, use that for pop fraction
            numerator = filter_stats["selected"]
            denominator = numerator + filter_stats["other"]
        else:
            # If new format is not available, default to old-style calculation
            numerator = self._cube_dict["result"].get("filtered", {}).get("weighted_n")
            denominator = (
                self._cube_dict["result"].get("unfiltered", {}).get("weighted_n")
            )

        try:
            return numerator / denominator
        except ZeroDivisionError:
            return np.nan
        except Exception:
            return 1.0

    @lazyproperty
    def unweighted_counts(self):
        """_UnweightedCountMeasure object for this cube.

        This object provides access to unweighted counts for this cube,
        whether or not the cube contains weighted counts.
        """
        return _UnweightedCountMeasure(self._cube_dict, self._all_dimensions)

    @lazyproperty
    def weighted_counts(self):
        """_WeightedCountMeasure object for this cube.

        This object provides access to weighted counts for this cube, if
        available. If the cube response is not weighted, the
        _UnweightedCountMeasure object for this cube is returned.
        """
        return (
            _WeightedCountMeasure(self._cube_dict, self._all_dimensions)
            if self.is_weighted
            else _UnweightedCountMeasure(self._cube_dict, self._all_dimensions)
        )


class _BaseMeasure(object):
    """Base class for measure objects."""

    def __init__(self, cube_dict, all_dimensions):
        self._cube_dict = cube_dict
        self._all_dimensions = all_dimensions

    @lazyproperty
    def raw_cube_array(self):
        """Return read-only ndarray of measure values from cube-response.

        The shape of the ndarray mirrors the shape of the (raw) cube
        response. Specifically, it includes values for missing elements, any
        MR_CAT dimensions, and any prunable rows and columns.
        """
        array = np.array(self._flat_values).reshape(self._all_dimensions.shape)
        # ---must be read-only to avoid hard-to-find bugs---
        array.flags.writeable = False
        return array

    @lazyproperty
    def _flat_values(self):  # pragma: no cover
        """Return tuple of mean values as found in cube response.

        This property must be implemented by each subclass.
        """
        raise NotImplementedError("must be implemented by each subclass")


class _MeanMeasure(_BaseMeasure):
    """Statistical mean values from a cube-response."""

    @lazyproperty
    def missing_count(self):
        """numeric representing count of missing rows reflected in response."""
        return self._cube_dict["result"]["measures"]["mean"].get("n_missing", 0)

    @lazyproperty
    def _flat_values(self):
        """Return tuple of mean values as found in cube response.

        Mean data may include missing items represented by a dict like
        {'?': -1} in the cube response. These are replaced by np.nan in the
        returned value.
        """
        return tuple(
            np.nan if type(x) is dict else x
            for x in self._cube_dict["result"]["measures"]["mean"]["data"]
        )


class _UnweightedCountMeasure(_BaseMeasure):
    """Unweighted counts for cube."""

    @lazyproperty
    def _flat_values(self):
        """tuple of int counts before weighting."""
        return tuple(self._cube_dict["result"]["counts"])


class _WeightedCountMeasure(_BaseMeasure):
    """Weighted counts for cube."""

    @lazyproperty
    def _flat_values(self):
        """tuple of numeric counts after weighting."""
        return tuple(self._cube_dict["result"]["measures"]["count"]["data"])



# encoding: utf-8

"""A stripe is a 1D data partition, roughly a one-dimension version of matrix.

Each strand object is based on a stripe, which is where most of its values come from.
Unlike a matrix, a stripe does not represent a crosstab and does not have "body" values.
It does however have summary values of some of the same types, like counts and
proportions. It does *not* have pvals or zscores.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import collections

import numpy as np

from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.util import lazyproperty


class TransformedStripe(object):
    """Stripe that reflects application of all transforms.

    `rows_dimension` is a `cr.cube.dimension.Dimension` object representing the single
    dimension of this stripe. This is referred to as the "rows" dimension because when
    viewed as a "card" in the Crunch UI, the elements of this dimension each appear as
    a distinct row of the card.

    `base_stripe` is one of the low-level, pre-transformation stripe objects defined
    later in the module. The `.stripe()` classmethod uses a factory method to construct
    an instance of the right class depending on the dimension type.

    This object is not meant to be constructed directly, except perhaps in unit tests.
    Use the `.stripe()` classmethod to create a new `TransformedStripe` instance.
    """

    def __init__(self, rows_dimension, base_stripe):
        self._rows_dimension = rows_dimension
        self._base_stripe = base_stripe

    @classmethod
    def stripe(cls, cube, rows_dimension, ca_as_0th, slice_idx):
        """Return a TransformedStripe object constructed from the data in `cube`."""
        base_stripe = _BaseBaseStripe.factory(
            cube, rows_dimension, ca_as_0th, slice_idx
        )
        return cls(rows_dimension, base_stripe)

    @lazyproperty
    def rows(self):
        """Sequence of post-transformation row vectors."""
        return tuple(row for row in self.rows_including_hidden if not row.hidden)

    @lazyproperty
    def rows_including_hidden(self):
        """Sequence of row vectors including those hidden by the user."""
        # --- ordering and insertion transforms are applied here. `._ordered_rows`
        # --- applies any ordering transforms and _StripeInsertionHelper creates and
        # --- interleaves subtotal rows.
        return tuple(
            _StripeInsertionHelper.iter_interleaved_rows(
                self._rows_dimension, self._ordered_rows, self._table_margin
            )
        )

    @lazyproperty
    def table_base_unpruned(self):
        """np.int64 scalar or 1D np.int64 ndarray of table base including "empty" items.

        Value is an np.int64 scalar except when this stripe is multiple-response (MR).
        This value is a 1D ndarray of np.int64 for an MR stripe.
        """
        # TODO: This the name is misleading. It's not only "unpruned" it's
        # "before_hiding" (of either kind, prune or hide-transform). But the real
        # problem is having this interface property at all. The need for this is related
        # to expressing ranges for base and margin in cubes that have an MR dimension.
        # The real solution is to compute ranges in `cr.cube` rather than leaking this
        # sort of internal detail through the interface and making the client compute
        # those for themselves. So this will require reconstructing that "show-ranges"
        # requirement and either adding some sort of a `.range` property that returns
        # a sequence of (min, max) tuples, or maybe just returning margin or base as
        # tuples when appropriate and having something like a `.margin_is_ranges`
        # predicate the client can switch on to control their rendering.
        return self._base_stripe.table_base

    @lazyproperty
    def table_margin_unpruned(self):
        """np.float/int64 scalar or 1D np.float/int64 ndarray of table margin.

        This value includes rows that may not appear in `.rows` because they are hidden.

        This value is an np.int64 scalar when this stripe is CAT. The value is an array
        when this stripe is multiple-response (MR) because each subvar of an MR
        dimension has a distinct table margin.

        Values are np.int64 when the cube-result is unweighted.
        """
        # TODO: see TODO in `.table_base_unpruned`
        return self._base_stripe.table_margin

    @lazyproperty
    def _ordered_rows(self):
        """Sequence of base-rows in order specified by user.

        Default ordering is "payload order", the order dimension elements appear in the
        cube result. This order may be overridden by an order transform.
        """
        # --- display_order is like (2, 1, 0, 3); each int item is the offset of a row
        # --- in the base-rows collection.
        rows = self._base_stripe.rows
        return tuple(rows[row_idx] for row_idx in self._rows_dimension.display_order)

    @lazyproperty
    def _table_margin(self):
        """Needed by inserted rows."""
        return self._base_stripe.table_margin


class _StripeInsertionHelper(object):
    """Base class for ordering partitions."""

    def __init__(self, rows_dimension, ordered_rows, table_margin):
        self._rows_dimension = rows_dimension
        self._ordered_rows = ordered_rows
        self._table_margin = table_margin

    @classmethod
    def iter_interleaved_rows(cls, rows_dimension, ordered_rows, table_margin):
        """Generate rows with subtotals inserted in correct position."""
        return cls(rows_dimension, ordered_rows, table_margin)._iter_interleaved_rows()

    def _iter_interleaved_rows(self):
        """Generate all row vectors with inserted rows interleaved at right spot."""
        # --- organize inserted-rows by anchor ---
        inserted_rows_by_anchor = collections.defaultdict(list)
        for inserted_row in self._inserted_rows:
            inserted_rows_by_anchor[inserted_row.anchor].append(inserted_row)

        # --- subtotals inserted at top ---
        for inserted_row in inserted_rows_by_anchor["top"]:
            yield inserted_row

        # --- body rows with subtotals anchored to specific body positions ---
        for row in self._ordered_rows:
            yield row
            for inserted_row in inserted_rows_by_anchor[row.element_id]:
                yield inserted_row

        # --- subtotals appended at bottom ---
        for inserted_row in inserted_rows_by_anchor["bottom"]:
            yield inserted_row

    @lazyproperty
    def _inserted_rows(self):
        """Sequence of _StripeInsertedRow objects representing inserted subtotal rows.

        The returned vectors are in the order subtotals were specified in the cube
        result, which is no particular order.
        """
        # ---an aggregate rows-dimension is not summable---
        if self._rows_dimension.dimension_type in (DT.MR, DT.CA):
            return tuple()

        return tuple(
            _StripeInsertedRow(subtotal, self._ordered_rows, self._table_margin)
            for subtotal in self._rows_dimension.subtotals
        )


# ===BASE STRIPES===


class _BaseBaseStripe(object):
    """Base class for all stripe objects."""

    def __init__(self, rows_dimension, unweighted_counts):
        self._rows_dimension = rows_dimension
        self._unweighted_counts = unweighted_counts

    @classmethod
    def factory(cls, cube, rows_dimension, ca_as_0th, slice_idx):
        """Return a base-slice object of appropriate type based on parameters."""
        counts = cube.counts
        unweighted_counts = cube.unweighted_counts

        # ---for cubes with means, create one of the means-stripe types---
        if cube.has_means:
            if rows_dimension.dimension_type == DT.MR:
                return _MeansWithMrStripe(rows_dimension, counts, unweighted_counts)
            return _MeansStripe(rows_dimension, counts, unweighted_counts)

        if ca_as_0th:
            return _CatStripe(
                rows_dimension, counts[slice_idx], unweighted_counts[slice_idx]
            )

        if rows_dimension.dimension_type == DT.MR:
            return _MrStripe(rows_dimension, counts, unweighted_counts)

        return _CatStripe(rows_dimension, counts, unweighted_counts)

    @lazyproperty
    def rows(self):
        """Sequence of rows in this stripe.

        The sequence includes a row for each valid element in the rows-dimension, in the
        order those elements are defined in the dimension (which is also the order in
        which that dimension's values appear in the cube result).
        """
        raise NotImplementedError(
            "must be implemented by each subclass"
        )  # pragma: no cover

    @lazyproperty
    def table_base(self):
        """np.int64 count of actual respondents asked this question."""
        return np.sum(self._unweighted_counts)

    @lazyproperty
    def _row_elements(self):
        """cr.cube.dimension._ValidElements instance for rows-dimension."""
        return self._rows_dimension.valid_elements


class _CatStripe(_BaseBaseStripe):
    """Special case of CAT x CAT, where the 2nd CAT doesn't exist.

    Values are treated as rows, while there's only a single column (vector).
    """

    def __init__(self, rows_dimension, counts, unweighted_counts):
        super(_CatStripe, self).__init__(rows_dimension, unweighted_counts)
        self._counts = counts

    @lazyproperty
    def rows(self):
        """tuple of _CatStripeRow for each valid element in rows-dimension."""
        table_margin = self.table_margin
        return tuple(
            _CatStripeRow(element, count, unweighted_count, table_margin)
            for (element, count, unweighted_count) in zip(
                self._row_elements, self._counts, self._unweighted_counts
            )
        )

    @lazyproperty
    def table_margin(self):
        """np.int64 total weighted count of responses to this question.

        Needed by inserted rows in later transformation step.
        """
        return np.sum(self._counts)


class _MeansStripe(_BaseBaseStripe):
    """A 1D calculator for a strand containing mean first-order measure."""

    def __init__(self, rows_dimension, means, unweighted_counts):
        super(_MeansStripe, self).__init__(rows_dimension, unweighted_counts)
        self._means = means

    @lazyproperty
    def rows(self):
        """Sequence of _MeansStripeRow for each valid element in rows-dimension."""
        return tuple(
            _MeansStripeRow(element, unweighted_count, mean)
            for element, unweighted_count, mean in zip(
                self._row_elements, self._unweighted_counts, self._means
            )
        )

    @lazyproperty
    def table_margin(self):
        # TODO: explain in docstring why this is unweighted for means instead of _counts
        return np.sum(self._unweighted_counts)


class _MeansWithMrStripe(_BaseBaseStripe):
    """Means behavior differs when dimension is MR."""

    def __init__(self, rows_dimension, means, unweighted_counts):
        super(_MeansWithMrStripe, self).__init__(rows_dimension, unweighted_counts)
        self._means = means

    @lazyproperty
    def rows(self):
        """Sequence of _MeansMrStripeRow for each valid element in rows-dimension."""
        return tuple(
            _MeansMrStripeRow(element, unweighted_counts, means)
            for element, unweighted_counts, means in zip(
                self._row_elements, self._unweighted_counts, self._means
            )
        )

    @lazyproperty
    def table_margin(self):
        # TODO: explain in docstring why this is unweighted for means instead of _counts
        return np.sum(self._unweighted_counts)


class _MrStripe(_BaseBaseStripe):
    """Special case of 1D MR slice (stripe)."""

    def __init__(self, rows_dimension, counts, unweighted_counts):
        super(_MrStripe, self).__init__(rows_dimension, unweighted_counts)
        self._counts = counts

    @lazyproperty
    def rows(self):
        """Sequence of _MrStripeRow for each valid element in rows-dimension."""
        return tuple(
            _MrStripeRow(element, counts, unweighted_counts, table_margin)
            for (element, counts, unweighted_counts, table_margin) in zip(
                self._row_elements,
                self._counts,
                self._unweighted_counts,
                self.table_margin,
            )
        )

    @lazyproperty
    def table_base(self):
        """1D ndarray of unweighted selected + unselected count for each row of stripe.

        Note the base (how many respondents were asked the question) can vary from row
        to row because not all choices are necessarily shown to all respondents.
        """
        return np.sum(self._unweighted_counts, axis=1)

    @lazyproperty
    def table_margin(self):
        """1D ndarray of weighted selected + unselected count for each row of stripe.

        Note this can vary from row to row because not all choices are necessarily shown
        to all respondents.
        """
        return np.sum(self._counts, axis=1)


# ===STRIPE ROWS===


class _StripeInsertedRow(object):
    """Represents an inserted (subtotal) row.

    This row item participates like any other row item, and a lot of its public
    properties are used by the strand object to form aggregate values across all rows.
    """

    def __init__(self, subtotal, base_rows, table_margin):
        self._subtotal = subtotal
        self._base_rows = base_rows
        self._table_margin = table_margin

    @lazyproperty
    def anchor(self):
        """str or int anchor value for this inserted.

        Value can be "top", "bottom" or an int element-id indicating the position of
        this inserted-row with respect to the base-rows.
        """
        return self._subtotal.anchor

    @lazyproperty
    def base(self):
        """np.int64 sum of (unweighted) sample population of each addend row."""
        return sum(row.base for row in self._addend_rows)

    @lazyproperty
    def count(self):
        """np.float64 subtotal of weighted count of this answer across addend rows."""
        return sum(row.count for row in self._addend_rows)

    @lazyproperty
    def fill(self):
        """An inserted row can have no element-fill-color transform."""
        return None

    @lazyproperty
    def hidden(self):
        """True if subtotal is pruned. Unconditionally False for stripe subtotal row."""
        return False

    @lazyproperty
    def is_inserted(self):
        """True when this row is an inserted row.

        Unconditionally True for _StripeInsertedRow.
        """
        return True

    @lazyproperty
    def label(self):
        """str heading for this row."""
        return self._subtotal.label

    @lazyproperty
    def mean(self):
        """Means cannot be subtotaled (so far at least)."""
        return np.nan

    @lazyproperty
    def numeric_value(self):
        """Numeric value (or np.nan) assigned to the element of this row.

        This is unconditionally np.nan for a subtotal row.
        """
        return np.nan

    @lazyproperty
    def table_margin(self):
        return self._table_margin

    @lazyproperty
    def table_proportions(self):
        """float 0.0->1.0 portion of weighted respondents who chose this response."""
        return self.count / self.table_margin

    @lazyproperty
    def unweighted_count(self):
        """int64 sum of unweighted-count of each addend row."""
        return sum(row.unweighted_count for row in self._addend_rows)

    @lazyproperty
    def _addend_rows(self):
        """Sequence of _BaseStripeRow subclass that contribute to this subtotal."""
        return tuple(
            row
            for row in self._base_rows
            if row.element_id in self._subtotal.addend_ids
        )


class _BaseStripeRow(object):
    """Base class for all stripe-row objects.

    Provides attributes drawn from the row element.
    """

    def __init__(self, element):
        self._element = element

    @lazyproperty
    def element_id(self):
        """int identifier of category or subvariable this row represents."""
        return self._element.element_id

    @lazyproperty
    def fill(self):
        """str RGB color like "#def032" or None when not specified.

        The value reflects the resolved element-fill transform cascade. A value of
        `None` indicates no element-fill transform was specified and the default
        (theme-specified) color should be used for this element.
        """
        return self._element.fill

    @lazyproperty
    def hidden(self):
        """True if vector is hidden.

        Vectors are hidden in two ways:

            1. Explicitly via transforms
            2. Implicitly when the base is 0 (also called pruning)

        This property checks whether a vector needs to be hidden, either implicitly or
        explicitly. It is used when iterating through rows or columns, to form the
        correct result.
        """
        return self._element.is_hidden or (self._element.prune and self.pruned)

    @lazyproperty
    def is_inserted(self):
        return False

    @lazyproperty
    def label(self):
        return self._element.label

    @lazyproperty
    def mean(self):
        """Default mean value is `np.nan`, consistent with an inserted row.

        It would be very unusual for this property to be accessed on a cube with no mean
        values, but this prevents raising an AttributeError and provides a reasonable
        value.
        """
        return np.nan

    @lazyproperty
    def numeric_value(self):
        """Numeric value (or np.nan) assigned to the element of this row.

        This allows a categorical value to participate in certain quantitative
        representations. For example, very-x to not-at-all-x could be assigned the
        values from 5 to 1, allowing a mean numeric sentiment like "3.5 on scale of 5"
        to be calculated.

        This value is np.nan when no numeric value has been assigned or numeric value is
        not applicable (such as for a numeric variable).
        """
        return self._element.numeric_value

    @lazyproperty
    def unweighted_count(self):
        """np.int64 unweighted-count for this row."""
        raise NotImplementedError(
            "`.unweighted_count` must be implemented by each subclass"
        )  # pragma: no cover


class _CatStripeRow(_BaseStripeRow):
    """Stripe-row for a categorical rows-dimension."""

    def __init__(self, element, count, unweighted_count, table_margin):
        super(_CatStripeRow, self).__init__(element)
        # ---count is a numpy numeric (np.int64 or np.float64)---
        self._count = count
        # ---unweighted_count is np.int64---
        self._unweighted_count = unweighted_count
        # ---table_margin is np.int64 or np.float64---
        self._table_margin = table_margin

    @lazyproperty
    def base(self):
        """np.int64 count of actual respondents who were offered this response."""
        # TODO: this doesn't seem right. The unweighted-count is the number of actual
        # respondents who *selected* this response, not all those who were asked the
        # question (as it should be).
        return self._unweighted_count

    @lazyproperty
    def count(self):
        """Weighted count for this row."""
        return self._count

    @lazyproperty
    def pruned(self):
        """True if this question was answered by exactly zero respondents."""
        return self._unweighted_count == 0 or np.isnan(self._unweighted_count)

    @lazyproperty
    def table_proportions(self):
        """float between 0.0 and 1.0 quantifying this row's share of stripe total.

        This is the proportion of *weighted* respondents who selected this response to
        the *weighted total* of respondents asked this question.
        """
        return self.count / self._table_margin

    @lazyproperty
    def unweighted_count(self):
        """np.int64 unweighted count for this row."""
        return self._unweighted_count


class _MeansStripeRow(_BaseStripeRow):
    """Stripe-row for a non-MR 1D cube-result with means."""

    def __init__(self, element, unweighted_count, mean):
        super(_MeansStripeRow, self).__init__(element)
        # ---unweighted_count is np.int64---
        self._unweighted_count = unweighted_count
        # ---mean is np.float64 or np.nan---
        self._mean = mean

    @lazyproperty
    def base(self):
        """np.int64 base, just unweighted count for a non-MR row."""
        return self._unweighted_count

    @lazyproperty
    def count(self):
        """It's a programming error if this is ever called."""
        raise NotImplementedError(
            "Mean measure has no unweighted count."
        )  # pragma: no cover

    @lazyproperty
    def mean(self):
        return self._mean

    @lazyproperty
    def pruned(self):
        return self._unweighted_count == 0 or np.isnan(self._unweighted_count)

    @lazyproperty
    def unweighted_count(self):
        """np.int64 unweighted-count for this row."""
        return self._unweighted_count


class _MeansMrStripeRow(_BaseStripeRow):
    """Stripe-row for a univariate MR cube-response with means."""

    def __init__(self, element, unweighted_counts, means):
        super(_MeansMrStripeRow, self).__init__(element)
        # ---unweighted_counts is a [selected, not] int array like [336 5501]---
        self._unweighted_counts = unweighted_counts
        # ---means is a [selected, not] float array like [2.578 1.639]---
        self._means = means

    @lazyproperty
    def base(self):
        """np.int64 unweighted count of selected."""
        return self._unweighted_counts[0]

    @lazyproperty
    def mean(self):
        """float mean of selected."""
        return self._means[0]

    @lazyproperty
    def pruned(self):
        return self.base == 0 or np.isnan(self.base)

    @lazyproperty
    def table_base(self):
        return self.base


class _MrStripeRow(_BaseStripeRow):
    """Stripe-row for use in MR stripe."""

    def __init__(self, element, counts, unweighted_counts, table_margin):
        super(_MrStripeRow, self).__init__(element)
        # --- counts is a float [selected, not-selected] array like [42.0 63.5] ---
        self._counts = counts
        # --- unweighted_counts is [selected, not-selected] int array like [607 143]---
        self._unweighted_counts = unweighted_counts
        # --- table_margin is a numpy numeric scalar ---
        self._table_margin = table_margin

    @lazyproperty
    def count(self):
        """np.float/int64 number of weighted respondents who selected this response."""
        return self._counts[0]

    @lazyproperty
    def pruned(self):
        """True if this question was asked of exactly zero actual respondents."""
        return self.table_base == 0

    @lazyproperty
    def table_base(self):
        """np.int64 sample size for MR question, how many actual folks were asked."""
        return np.sum(self._both_unweighted_counts)

    @lazyproperty
    def table_margin(self):
        """np.float64 weighted sample size, how many "weighted" folks were asked."""
        return np.sum(self._both_counts)

    @lazyproperty
    def table_proportions(self):
        """np.float64 proportion of all weighted respondents selecteing this answer."""
        return self.count / self._table_margin

    @lazyproperty
    def unweighted_count(self):
        """np.int64 unweighted count of selected responses."""
        return self._unweighted_counts[0]

    @lazyproperty
    def _both_counts(self):
        """np.float64 count of weighted respondents offered this possible response."""
        return self._counts

    @lazyproperty
    def _both_unweighted_counts(self):
        """np.int64 count of actual respondents offered this possible response."""
        return self._unweighted_counts



# encoding: utf-8

"""Unit test suite for cr.cube.dimension module."""

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import pytest

from cr.cube.dimension import (
    AllDimensions,
    _AllElements,
    _ApparentDimensions,
    _BaseDimensions,
    _BaseElements,
    Dimension,
    _DimensionFactory,
    _Element,
    _ElementTransforms,
    _RawDimension,
    _Subtotal,
    _Subtotals,
    _ValidElements,
)
from cr.cube.enums import DIMENSION_TYPE as DT

from ..unitutil import (
    ANY,
    call,
    class_mock,
    initializer_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class Describe_BaseDimensions(object):
    def it_has_sequence_behaviors(self, request, _dimensions_prop_):
        _dimensions_prop_.return_value = (0, 1, 2)
        base_dimensions = _BaseDimensions()

        assert base_dimensions[1] == 1
        assert base_dimensions[1:3] == (1, 2)
        assert len(base_dimensions) == 3
        assert list(n for n in base_dimensions) == [0, 1, 2]

    def it_stores_its_dimension_objects_in_a_tuple_to_help(self):
        base_dimensions = _BaseDimensions()
        with pytest.raises(NotImplementedError):
            base_dimensions._dimensions

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _dimensions_prop_(self, request):
        return property_mock(request, _BaseDimensions, "_dimensions")


class DescribeAllDimensions(object):
    """Unit-test suite for `cr.cube.dimension.AllDimensions` object."""

    def it_provides_access_to_its_ApparentDimensions(
        self,
        _ApparentDimensions_,
        apparent_dimensions_,
        _dimensions_prop_,
        all_dimensions_,
    ):
        _dimensions_prop_.return_value = all_dimensions_
        _ApparentDimensions_.return_value = apparent_dimensions_
        all_dimensions = AllDimensions(None)

        apparent_dimensions = all_dimensions.apparent_dimensions

        _ApparentDimensions_.assert_called_once_with(all_dimensions=all_dimensions_)
        assert apparent_dimensions is apparent_dimensions_

    def it_knows_its_shape(self, request, _dimensions_prop_):
        _dimensions_prop_.return_value = tuple(
            instance_mock(request, Dimension, name="dim-%d" % idx, shape=element_count)
            for idx, element_count in enumerate((3, 2, 1))
        )
        all_dimensions = AllDimensions(None)

        shape = all_dimensions.shape

        assert shape == (3, 2, 1)

    def it_stores_its_dimensions_in_a_tuple_to_help(self, request, _DimensionFactory_):
        dimension_dicts_ = [{"d": 0}, {"d": 1}, {"d": 2}]
        dimensions_ = tuple(
            instance_mock(request, Dimension, name="dim-%d" % idx) for idx in range(3)
        )
        _DimensionFactory_.iter_dimensions.return_value = iter(dimensions_)
        all_dimensions = AllDimensions(dimension_dicts_)

        dimensions = all_dimensions._dimensions

        _DimensionFactory_.iter_dimensions.assert_called_once_with(dimension_dicts_)
        assert dimensions == dimensions_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def all_dimensions_(self, request):
        return instance_mock(request, AllDimensions)

    @pytest.fixture
    def _ApparentDimensions_(self, request):
        return class_mock(request, "cr.cube.dimension._ApparentDimensions")

    @pytest.fixture
    def apparent_dimensions_(self, request):
        return instance_mock(request, _ApparentDimensions)

    @pytest.fixture
    def _DimensionFactory_(self, request):
        return class_mock(request, "cr.cube.dimension._DimensionFactory")

    @pytest.fixture
    def _dimensions_prop_(self, request):
        return property_mock(request, AllDimensions, "_dimensions")


class Describe_ApparentDimensions(object):
    def it_stores_its_dimensions_in_a_tuple_to_help(self, request):
        all_dimensions_ = tuple(
            instance_mock(request, Dimension, name="dim-%d" % idx, dimension_type=dt)
            for idx, dt in enumerate((DT.CAT, DT.MR, DT.MR_CAT))
        )
        apparent_dimensions = _ApparentDimensions(all_dimensions_)

        dimensions = apparent_dimensions._dimensions

        assert dimensions == all_dimensions_[:2]


class Describe_DimensionFactory(object):
    def it_provides_an_interface_classmethod(
        self, request, dimension_dicts_, _init_, _iter_dimensions_
    ):
        dimensions_ = tuple(
            instance_mock(request, Dimension, name="dim-%d" % idx) for idx in range(3)
        )
        _iter_dimensions_.return_value = iter(dimensions_)

        dimension_iter = _DimensionFactory.iter_dimensions(dimension_dicts_)

        # ---ANY is for dimension_factory object (self), which we can't see---
        _init_.assert_called_once_with(ANY, dimension_dicts_)
        _iter_dimensions_.assert_called_once_with(ANY)
        assert tuple(dimension_iter) == dimensions_

    def it_generates_the_dimensions_for_a_cube(
        self, request, dimension_dicts_, _raw_dimensions_prop_, Dimension_
    ):
        dimension_types_ = (DT.CAT, DT.CA_SUBVAR, DT.CA_CAT)
        _raw_dimensions_prop_.return_value = tuple(
            instance_mock(
                request,
                _RawDimension,
                name="raw-dim-%d" % idx,
                dimension_dict=dimension_dicts_[idx],
                dimension_type=dimension_types_[idx],
            )
            for idx in range(3)
        )
        dimensions_ = tuple(
            instance_mock(request, Dimension, name="dim-%d" % idx) for idx in range(3)
        )
        Dimension_.side_effect = iter(dimensions_)
        dimension_factory = _DimensionFactory(None)

        dimension_iter = dimension_factory._iter_dimensions()

        # ---exercising the iterator needs to come first---
        assert tuple(dimension_iter) == dimensions_
        assert Dimension_.call_args_list == [
            call(dimension_dicts_[0], dimension_types_[0]),
            call(dimension_dicts_[1], dimension_types_[1]),
            call(dimension_dicts_[2], dimension_types_[2]),
        ]

    def it_constructs_RawDimension_objects_to_help(
        self, request, dimension_dicts_, _RawDimension_
    ):
        raw_dimensions_ = tuple(
            instance_mock(request, _RawDimension, name="raw-dim-%d" % idx)
            for idx in range(3)
        )
        _RawDimension_.side_effect = iter(raw_dimensions_)
        dimension_factory = _DimensionFactory(dimension_dicts_)

        raw_dimensions = dimension_factory._raw_dimensions

        assert _RawDimension_.call_args_list == [
            call(dimension_dicts_[0], dimension_dicts_),
            call(dimension_dicts_[1], dimension_dicts_),
            call(dimension_dicts_[2], dimension_dicts_),
        ]
        assert raw_dimensions == raw_dimensions_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def Dimension_(self, request):
        return class_mock(request, "cr.cube.dimension.Dimension")

    @pytest.fixture
    def dimension_dicts_(self, request):
        """Tuple of three shallow dimension-dict placeholders."""
        return ({"d": 0}, {"d": 1}, {"d": 2})

    @pytest.fixture
    def _init_(self, request):
        return initializer_mock(request, _DimensionFactory)

    @pytest.fixture
    def _iter_dimensions_(self, request):
        return method_mock(request, _DimensionFactory, "_iter_dimensions")

    @pytest.fixture
    def _RawDimension_(self, request):
        return class_mock(request, "cr.cube.dimension._RawDimension")

    @pytest.fixture
    def _raw_dimensions_prop_(self, request):
        return property_mock(request, _DimensionFactory, "_raw_dimensions")


class Describe_RawDimension(object):
    def it_provides_access_to_the_dimension_dict(self):
        dimension_dict_ = {"dimension": "dict"}
        raw_dimension = _RawDimension(dimension_dict_, None)

        dimension_dict = raw_dimension.dimension_dict

        assert dimension_dict == dimension_dict_

    @pytest.mark.parametrize(
        "dimension_dict, expected_value",
        (
            ({"type": {"class": "categorical"}}, "categorical"),
            (
                {"type": {"class": "enum", "subtype": {"class": "variable"}}},
                "enum.variable",
            ),
        ),
    )
    def it_parses_the_base_type_to_help(self, dimension_dict, expected_value):
        raw_dimension = _RawDimension(dimension_dict, None)

        base_type = raw_dimension._base_type

        assert base_type == expected_value

    def but_it_raises_on_unrecognized_type_class(self):
        raw_dimension = _RawDimension({"type": {"class": "crunched"}}, None)
        with pytest.raises(NotImplementedError):
            raw_dimension._base_type

    @pytest.mark.parametrize(
        "base_type, cat_type, arr_type, expected_value",
        (
            ("categorical", DT.CAT, None, DT.CAT),
            ("enum.variable", None, DT.MR, DT.MR),
            ("enum.datetime", None, None, DT.DATETIME),
            ("enum.numeric", None, None, DT.BINNED_NUMERIC),
            ("enum.text", None, None, DT.TEXT),
        ),
    )
    def it_determines_the_dimension_type(
        self,
        base_type,
        cat_type,
        arr_type,
        expected_value,
        _base_type_prop_,
        _resolve_categorical_,
        _resolve_array_type_,
    ):
        _base_type_prop_.return_value = base_type
        _resolve_categorical_.return_value = cat_type if cat_type else None
        _resolve_array_type_.return_value = arr_type if arr_type else None
        raw_dimension = _RawDimension(None, None)
        resolve_cat_calls = [call(raw_dimension)] if cat_type else []
        resolve_arr_calls = [call(raw_dimension)] if arr_type else []

        dimension_type = raw_dimension.dimension_type

        assert _resolve_categorical_.call_args_list == resolve_cat_calls
        assert _resolve_array_type_.call_args_list == resolve_arr_calls
        assert dimension_type == expected_value

    def but_it_raises_on_unrecognized_base_type(self, _base_type_prop_):
        raw_dimension = _RawDimension(None, None)
        _base_type_prop_.return_value = "hyper.dimensional"
        with pytest.raises(NotImplementedError):
            raw_dimension.dimension_type

    def it_knows_the_dimension_variable_identifier_to_help(self):
        dimension_dict = {"references": {"alias": "varski"}}
        raw_dimension = _RawDimension(dimension_dict, None)

        alias = raw_dimension._alias

        assert alias == "varski"

    @pytest.mark.parametrize(
        "dimension_dict, expected_value",
        (
            ({"type": {}}, False),
            ({"type": {"categories": []}}, False),
            ({"type": {"categories": [{}, {}]}}, False),
            ({"type": {"categories": [{"selected": False}, {}]}}, False),
            ({"type": {"categories": [{"selected": True}, {}]}}, True),
        ),
    )
    def it_can_tell_when_a_dimension_has_a_selected_category_to_help(
        self, dimension_dict, expected_value
    ):
        raw_dimension = _RawDimension(dimension_dict, None)

        has_selected_category = raw_dimension._has_selected_category

        assert has_selected_category is expected_value

    @pytest.mark.parametrize(
        "dimension_dict, expected_value",
        (({"references": {}}, False), ({"references": {"subreferences": {}}}, True)),
    )
    def it_distinguishes_an_array_categorical_type_to_help(
        self, dimension_dict, expected_value
    ):
        raw_dimension = _RawDimension(dimension_dict, None)

        is_array_cat = raw_dimension._is_array_cat

        assert is_array_cat == expected_value

    def it_finds_the_subsequent_raw_dimension_to_help(self, request, dimension_dicts_):
        dimension_dict_ = dimension_dicts_[1]
        next_dimension_dict_ = dimension_dicts_[2]
        raw_dimension = _RawDimension(dimension_dict_, dimension_dicts_)
        # --initializer must be mocked after contructing raw_dimension
        # --otherwise it would have no instance variables
        _init_ = initializer_mock(request, _RawDimension)

        next_raw_dimension = raw_dimension._next_raw_dimension

        _init_.assert_called_once_with(
            next_raw_dimension, next_dimension_dict_, dimension_dicts_
        )
        assert type(next_raw_dimension).__name__ == "_RawDimension"

    def but_it_returns_None_for_the_last_dimension(self, dimension_dicts_):
        dimension_dict_ = dimension_dicts_[2]
        raw_dimension = _RawDimension(dimension_dict_, dimension_dicts_)
        next_raw_dimension = raw_dimension._next_raw_dimension
        assert next_raw_dimension is None

    @pytest.mark.parametrize(
        "is_last, base_type, has_sel_cat, alias, expected_value",
        ((False, None, None, None, DT.CA),),
    )
    def it_resolves_an_array_type_to_help(
        self,
        is_last,
        base_type,
        has_sel_cat,
        alias,
        expected_value,
        raw_dimension_,
        _next_raw_dimension_prop_,
    ):
        raw_dimension_._base_type = base_type
        raw_dimension_._has_selected_category = has_sel_cat
        raw_dimension_._alias = alias
        next_raw_dimension = None if is_last else raw_dimension_
        _next_raw_dimension_prop_.return_value = next_raw_dimension
        raw_dimension = _RawDimension(None, None)

        dimension_type = raw_dimension._resolve_array_type()

        assert dimension_type == expected_value

    @pytest.mark.parametrize(
        "is_array_cat, has_selected_cat, expected_value",
        (
            (False, False, DT.CAT),
            (False, True, DT.LOGICAL),
            (True, False, DT.CA_CAT),
            (True, True, DT.MR_CAT),
        ),
    )
    def it_resolves_a_categorical_type_to_help(
        self,
        is_array_cat,
        has_selected_cat,
        expected_value,
        _is_array_cat_prop_,
        _has_selected_category_prop_,
    ):
        _is_array_cat_prop_.return_value = is_array_cat
        _has_selected_category_prop_.return_value = has_selected_cat
        raw_dimension = _RawDimension(None, None)

        dimension_type = raw_dimension._resolve_categorical()

        assert dimension_type == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _base_type_prop_(self, request):
        return property_mock(request, _RawDimension, "_base_type")

    @pytest.fixture
    def dimension_dicts_(self):
        return ({"dim", 0}, {"dim", 1}, {"dim", 2})

    @pytest.fixture
    def _has_selected_category_prop_(self, request):
        return property_mock(request, _RawDimension, "_has_selected_category")

    @pytest.fixture
    def _is_array_cat_prop_(self, request):
        return property_mock(request, _RawDimension, "_is_array_cat")

    @pytest.fixture
    def _next_raw_dimension_prop_(self, request):
        return property_mock(request, _RawDimension, "_next_raw_dimension")

    @pytest.fixture
    def raw_dimension_(self, request):
        return instance_mock(request, _RawDimension)

    @pytest.fixture
    def _resolve_array_type_(self, request):
        return method_mock(request, _RawDimension, "_resolve_array_type")

    @pytest.fixture
    def _resolve_categorical_(self, request):
        return method_mock(request, _RawDimension, "_resolve_categorical")


class DescribeDimension(object):
    """Unit-test suite for `cr.cube.dimension.Dimension` object."""

    @pytest.mark.parametrize(
        "dimension_dict, expected_value",
        (
            ({"references": {}}, ""),
            ({"references": {"description": None}}, ""),
            ({"references": {"description": ""}}, ""),
            ({"references": {"description": "Crunchiness"}}, "Crunchiness"),
        ),
    )
    def it_knows_its_description(self, dimension_dict, expected_value):
        dimension = Dimension(dimension_dict, None)

        description = dimension.description

        assert description == expected_value

    def it_knows_its_dimension_type(self):
        dimension = Dimension(None, DT.CAT)
        dimension_type = dimension.dimension_type
        assert dimension_type == DT.CAT

    def it_knows_the_numeric_values_of_its_elements(
        self, request, valid_elements_prop_
    ):
        valid_elements_prop_.return_value = tuple(
            instance_mock(request, _Element, numeric_value=numeric_value)
            for numeric_value in (1, 2.2, np.nan)
        )
        dimension = Dimension(None, None)

        numeric_values = dimension.numeric_values

        assert numeric_values == (1, 2.2, np.nan)

    @pytest.mark.parametrize(
        "dimension_dict, insertion_dicts",
        (
            ({}, []),
            ({"references": {}}, []),
            ({"references": {"view": {}}}, []),
            ({"references": {"view": {"transform": {}}}}, []),
            ({"references": {"view": {"transform": {"insertions": []}}}}, []),
            (
                {
                    "references": {
                        "view": {
                            "transform": {
                                "insertions": [
                                    {"insertion": "dict-1"},
                                    {"insertion": "dict-2"},
                                ]
                            }
                        }
                    }
                },
                [{"insertion": "dict-1"}, {"insertion": "dict-2"}],
            ),
        ),
    )
    def it_provides_access_to_its_subtotals_to_help(
        self,
        dimension_dict,
        insertion_dicts,
        _Subtotals_,
        subtotals_,
        valid_elements_prop_,
        valid_elements_,
        prune_prop_,
    ):
        dimension_dict["type"] = {"class": "categorical", "categories": []}
        valid_elements_prop_.return_value = valid_elements_
        _Subtotals_.return_value = subtotals_
        prune_prop_.return_value = True
        dimension = Dimension(dimension_dict, None)

        subtotals = dimension.subtotals

        _Subtotals_.assert_called_once_with(insertion_dicts, valid_elements_, True)
        assert subtotals is subtotals_

    def but_it_overrides_with_subtotal_transforms_when_present(
        self,
        valid_elements_prop_,
        valid_elements_,
        _Subtotals_,
        subtotals_,
        prune_prop_,
    ):
        dimension_transforms = {"insertions": ["subtotal", "dicts"]}
        valid_elements_prop_.return_value = valid_elements_
        _Subtotals_.return_value = subtotals_
        prune_prop_.return_value = False
        dimension = Dimension(None, None, dimension_transforms)

        subtotals = dimension.subtotals

        _Subtotals_.assert_called_once_with(
            ["subtotal", "dicts"], valid_elements_, False
        )
        assert subtotals is subtotals_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def prune_prop_(self, request):
        return property_mock(request, Dimension, "prune")

    @pytest.fixture
    def _Subtotals_(self, request):
        return class_mock(request, "cr.cube.dimension._Subtotals")

    @pytest.fixture
    def subtotals_(self, request):
        return instance_mock(request, _Subtotals)

    @pytest.fixture
    def valid_elements_(self, request):
        return instance_mock(request, _ValidElements)

    @pytest.fixture
    def valid_elements_prop_(self, request):
        return property_mock(request, Dimension, "valid_elements")


class Describe_BaseElements(object):
    """Unit-test suite for `cr.cube.dimension._BaseElements` object."""

    def it_has_sequence_behaviors(self, request, _elements_prop_):
        _elements_prop_.return_value = (1, 2, 3)
        elements = _BaseElements()

        assert elements[1] == 2
        assert elements[1:3] == (2, 3)
        assert len(elements) == 3
        assert list(n for n in elements) == [1, 2, 3]

    def it_knows_the_element_ids(self, request, _elements_prop_):
        _elements_prop_.return_value = tuple(
            instance_mock(request, _Element, element_id=n) for n in (1, 2, 5)
        )
        elements = _BaseElements()

        element_ids = elements.element_ids

        assert element_ids == (1, 2, 5)

    def it_knows_the_element_indices(self, request, _elements_prop_):
        _elements_prop_.return_value = tuple(
            instance_mock(request, _Element, index=index) for index in (1, 3, 4)
        )
        elements = _BaseElements()

        element_idxs = elements.element_idxs

        assert element_idxs == (1, 3, 4)

    def it_can_find_an_element_by_id(self, request, _elements_by_id_prop_):
        elements_ = tuple(
            instance_mock(request, _Element, element_id=element_id)
            for element_id in (3, 7, 11)
        )
        _elements_by_id_prop_.return_value = {
            element_.element_id: element_ for element_ in elements_
        }
        elements = _BaseElements()

        element = elements.get_by_id(7)

        assert element is elements_[1]

    def it_maintains_a_dict_of_elements_by_id_to_help(self, request, _elements_prop_):
        elements_ = tuple(
            instance_mock(request, _Element, element_id=element_id)
            for element_id in (4, 6, 7)
        )
        _elements_prop_.return_value = elements_
        elements = _BaseElements()

        elements_by_id = elements._elements_by_id

        assert elements_by_id == {4: elements_[0], 6: elements_[1], 7: elements_[2]}

    def it_stores_its_elements_in_a_tuple_to_help(self):
        base_elements = _BaseElements()
        # ---must be implemented by each subclass---
        with pytest.raises(NotImplementedError):
            base_elements._elements

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _elements_by_id_prop_(self, request):
        return property_mock(request, _BaseElements, "_elements_by_id")

    @pytest.fixture
    def _elements_prop_(self, request):
        return property_mock(request, _BaseElements, "_elements")


class Describe_AllElements(object):
    """Unit-test suite for `cr.cube.dimension._AllElements` object."""

    def it_provides_access_to_the_ValidElements_object(
        self, request, _elements_prop_, _ValidElements_, valid_elements_
    ):
        elements_ = tuple(
            instance_mock(request, _Element, name="el%s" % idx) for idx in range(3)
        )
        dimension_transforms_dict = {"dimension": "transforms"}
        _elements_prop_.return_value = elements_
        _ValidElements_.return_value = valid_elements_
        all_elements = _AllElements(None, dimension_transforms_dict)

        valid_elements = all_elements.valid_elements

        _ValidElements_.assert_called_once_with(elements_, dimension_transforms_dict)
        assert valid_elements is valid_elements_

    def it_creates_its_Element_objects_in_a_local_factory_to_help(
        self,
        request,
        _prune_prop_,
        _ElementTransforms_,
        _Element_,
        _iter_element_makings_,
    ):
        _prune_prop_.return_value = True
        element_transforms_ = tuple(
            instance_mock(request, _ElementTransforms, name="element-xfrms-%s" % idx)
            for idx in range(3)
        )
        _ElementTransforms_.side_effect = iter(element_transforms_)
        _iter_element_makings_.return_value = iter(
            (
                (0, {"element": "dict-A"}, {"xfrms": 0}),
                (1, {"element": "dict-B"}, {"xfrms": 1}),
                (2, {"element": "dict-C"}, {"xfrms": 2}),
            )
        )
        elements_ = tuple(
            instance_mock(request, _Element, name="element-%s" % idx)
            for idx in range(3)
        )
        _Element_.side_effect = iter(elements_)
        all_elements = _AllElements(None, None)

        elements = all_elements._elements

        assert _ElementTransforms_.call_args_list == [
            call({"xfrms": 0}, True),
            call({"xfrms": 1}, True),
            call({"xfrms": 2}, True),
        ]
        assert _Element_.call_args_list == [
            call({"element": "dict-A"}, 0, element_transforms_[0]),
            call({"element": "dict-B"}, 1, element_transforms_[1]),
            call({"element": "dict-C"}, 2, element_transforms_[2]),
        ]
        assert elements == (elements_[0], elements_[1], elements_[2])

    def it_generates_element_factory_inputs_to_help(self, _element_dicts_prop_):
        dimension_transforms_dict = {
            "elements": {"6": {"element": "xfrms_6"}, "4": {"element": "xfrms_4"}}
        }
        _element_dicts_prop_.return_value = (
            {"id": 4, "element": "dict_4"},
            {"id": 2, "element": "dict_2"},
            {"id": 6, "element": "dict_6"},
        )
        all_elements = _AllElements(None, dimension_transforms_dict)

        element_makings = tuple(all_elements._iter_element_makings())

        assert element_makings == (
            (0, {"id": 4, "element": "dict_4"}, {"element": "xfrms_4"}),
            (1, {"id": 2, "element": "dict_2"}, {}),
            (2, {"id": 6, "element": "dict_6"}, {"element": "xfrms_6"}),
        )

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _Element_(self, request):
        return class_mock(request, "cr.cube.dimension._Element")

    @pytest.fixture
    def _element_dicts_prop_(self, request):
        return property_mock(request, _AllElements, "_element_dicts")

    @pytest.fixture
    def _elements_prop_(self, request):
        return property_mock(request, _AllElements, "_elements")

    @pytest.fixture
    def _ElementTransforms_(self, request):
        return class_mock(request, "cr.cube.dimension._ElementTransforms")

    @pytest.fixture
    def _iter_element_makings_(self, request):
        return method_mock(request, _AllElements, "_iter_element_makings")

    @pytest.fixture
    def _prune_prop_(self, request):
        return property_mock(request, _AllElements, "_prune")

    @pytest.fixture
    def _ValidElements_(self, request):
        return class_mock(request, "cr.cube.dimension._ValidElements")

    @pytest.fixture
    def valid_elements_(self, request):
        return instance_mock(request, _ValidElements)


class Describe_ValidElements(object):
    """Unit-test suite for `cr.cube.dimension._ValidElements` object."""

    def it_gets_its_Element_objects_from_an_AllElements_object(
        self, request, all_elements_
    ):
        elements_ = tuple(
            instance_mock(request, _Element, name="element-%s" % idx, missing=missing)
            for idx, missing in enumerate([False, True, False])
        )
        all_elements_.__iter__.return_value = iter(elements_)
        valid_elements = _ValidElements(all_elements_, None)

        elements = valid_elements._elements

        assert elements == (elements_[0], elements_[2])

    @pytest.mark.parametrize(
        "explicit_order, expected_value ",
        ((None, (0, 1, 2)), ((1, 0, 4, 3), (1, 0, 4, 3))),
    )
    def it_knows_its_display_order(
        self, request, all_elements_, explicit_order, expected_value
    ):
        _explicit_order_ = property_mock(request, _ValidElements, "_explicit_order")
        elements_ = tuple(
            instance_mock(request, _Element, missing=False) for _ in range(3)
        )
        all_elements_.__iter__.return_value = iter(elements_)
        valid_elements = _ValidElements(all_elements_, None)
        _explicit_order_.return_value = explicit_order

        display_order = valid_elements.display_order

        assert display_order == expected_value

    @pytest.mark.parametrize(
        "element_transform_dict, expected_value",
        (
            ({}, None),
            ({"order": {"element_ids": [1, 3, 2], "type": "explicit"}}, (0, 2, 1)),
            ({"order": {"element_ids": [-1, 3, 2], "type": "explicit"}}, (2, 1, 0)),
            ({"order": {"element_ids": [1, 2, 3], "type": "explicit"}}, (0, 1, 2)),
        ),
    )
    def it_returns_the_correct_explicit_order(
        self, all_elements_, element_transform_dict, expected_value
    ):
        elements_ = tuple(_Element({"id": idx + 1}, None, None) for idx in range(3))
        all_elements_.__iter__.return_value = iter(elements_)
        valid_elements = _ValidElements(all_elements_, element_transform_dict)

        _explicit_order_ = valid_elements._explicit_order

        assert _explicit_order_ == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def all_elements_(self, request):
        return instance_mock(request, _AllElements)


class Describe_Element(object):
    """Unit-test suite for `cr.cube.dimension._Element` object."""

    def it_knows_its_element_id(self):
        element_dict = {"id": 42}
        element = _Element(element_dict, None, None)

        element_id = element.element_id

        assert element_id == 42

    def it_knows_its_fill_RGB_color_str(self, element_transforms_):
        element_transforms_.fill = [255, 255, 248]
        element = _Element(None, None, element_transforms_)

        rgb_color_fill = element.fill

        assert rgb_color_fill == [255, 255, 248]

    def it_knows_its_position_among_all_the_dimension_elements(self):
        element = _Element(None, 17, None)
        index = element.index
        assert index == 17

    # TODO: add test cases that exercise element-name transform
    @pytest.mark.parametrize(
        ("element_dict", "expected_value"),
        (
            ({}, ""),
            ({"name": ""}, ""),
            ({"name": None}, ""),
            ({"name": "Bob"}, "Bob"),
            ({"name": "Hinzufägen"}, "Hinzufägen"),
            ({"value": ["A", "F"]}, "A-F"),
            ({"value": [1.2, 3.4]}, "1.2-3.4"),
            ({"value": 42}, "42"),
            ({"value": 4.2}, "4.2"),
            ({"value": "Bill"}, "Bill"),
            ({"value": "Fähig"}, "Fähig"),
            ({"value": {"references": {}}}, ""),
            ({"value": {"references": {"name": "Tom"}}}, "Tom"),
        ),
    )
    def it_knows_its_label(self, element_dict, expected_value, element_transforms_):
        element_transforms_.name = None
        element = _Element(element_dict, None, element_transforms_)

        label = element.label

        assert label == expected_value

    @pytest.mark.parametrize(
        ("hide", "expected_value"), ((True, True), (False, False), (None, False))
    )
    def it_knows_whether_it_is_explicitly_hidden(self, request, hide, expected_value):
        element_transforms_ = instance_mock(request, _ElementTransforms, hide=hide)
        element = _Element(None, None, element_transforms_)

        is_hidden = element.is_hidden

        assert is_hidden is expected_value

    @pytest.mark.parametrize(
        ("element_dict", "expected_value"),
        (
            ({}, False),
            ({"missing": None}, False),
            ({"missing": False}, False),
            ({"missing": True}, True),
            # ---not expected values, but just in case---
            ({"missing": 0}, False),
            ({"missing": 1}, True),
        ),
    )
    def it_knows_whether_its_missing_or_valid(self, element_dict, expected_value):
        element = _Element(element_dict, None, None)

        missing = element.missing

        # ---only True or False, no Truthy or Falsy (so use `is` not `==`)---
        assert missing is expected_value

    @pytest.mark.parametrize(
        ("element_dict", "expected_value"),
        (
            ({}, np.nan),
            ({"numeric_value": None}, np.nan),
            ({"numeric_value": 0}, 0),
            ({"numeric_value": 7}, 7),
            ({"numeric_value": -3.2}, -3.2),
            # ---not expected values, just to document the behavior that
            # ---no attempt is made to convert values to numeric
            ({"numeric_value": "666"}, "666"),
            ({"numeric_value": {}}, {}),
            ({"numeric_value": {"?": 8}}, {"?": 8}),
        ),
    )
    def it_knows_its_numeric_value(self, element_dict, expected_value):
        element = _Element(element_dict, None, None)

        numeric_value = element.numeric_value

        # ---np.nan != np.nan, but np.nan in [np.nan] works---
        assert numeric_value in [expected_value]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def element_transforms_(self, request):
        return instance_mock(request, _ElementTransforms)


class DescribeElementTransforms(object):
    """Unit-test suite for `cr.cube.dimension._ElementTransforms` object."""

    @pytest.mark.parametrize(
        "element_transforms_dict, expected_value",
        (
            ({"name": "Zillow", "fill": "#316395"}, "#316395"),
            ({"name": "Zillow", "fill": None}, None),
            ({"name": "Zillow", "fill": [255, 255, 0]}, [255, 255, 0]),
        ),
    )
    def it_knows_its_fill_color_value(self, element_transforms_dict, expected_value):
        element_transforms = _ElementTransforms(element_transforms_dict, None)

        fill_color_value = element_transforms.fill

        assert fill_color_value == expected_value

    @pytest.mark.parametrize(
        "element_transforms_dict, expected_value",
        (
            ({"hide": True}, True),
            ({"hide": False}, False),
            ({"hide": None}, None),
            ({"hide": 0}, None),
        ),
    )
    def it_knows_when_it_is_explicitly_hidden(
        self, element_transforms_dict, expected_value
    ):
        element_transforms = _ElementTransforms(element_transforms_dict, None)

        is_hidden = element_transforms.hide

        assert is_hidden is expected_value

    @pytest.mark.parametrize(
        "element_transforms_dict, expected_value",
        (
            ({"name": "MyDisplayName"}, "MyDisplayName"),
            ({"name": ""}, ""),
            ({"name": None}, ""),
            ({"foo": "foo"}, None),
        ),
    )
    def it_knows_its_name(self, element_transforms_dict, expected_value):
        element_transforms = _ElementTransforms(element_transforms_dict, None)

        name = element_transforms.name

        assert name == expected_value


class Describe_Subtotals(object):
    """Unit-test suite for `cr.cube.dimension._Subtotals` object."""

    def it_has_sequence_behaviors(self, request, _subtotals_prop_):
        _subtotals_prop_.return_value = (1, 2, 3)
        subtotals = _Subtotals(None, None, None)

        assert subtotals[1] == 2
        assert subtotals[1:3] == (2, 3)
        assert len(subtotals) == 3
        assert list(n for n in subtotals) == [1, 2, 3]

    def it_can_iterate_subtotals_having_a_given_anchor(self, request, _subtotals_prop_):
        subtotals_ = tuple(
            instance_mock(request, _Subtotal, name="subtotal-%d" % idx, anchor=anchor)
            for idx, anchor in enumerate(["bottom", 2, "bottom"])
        )
        _subtotals_prop_.return_value = subtotals_
        subtotals = _Subtotals(None, None, None)

        subtotals_with_anchor = tuple(subtotals.iter_for_anchor("bottom"))

        assert subtotals_with_anchor == (subtotals_[0], subtotals[2])

    def it_provides_the_element_ids_as_a_set_to_help(self, request, valid_elements_):
        valid_elements_.element_ids = tuple(range(3))
        subtotals = _Subtotals(None, valid_elements_, None)

        element_ids = subtotals._element_ids

        assert element_ids == {0, 1, 2}

    @pytest.mark.parametrize(
        "insertion_dicts, element_ids, expected_value",
        (
            ([], (), ()),
            (["not-a-dict", None], (), ()),
            ([{"function": "hyperdrive"}], (), ()),
            ([{"function": "subtotal", "arghhs": []}], (), ()),
            ([{"function": "subtotal", "anchor": 9, "args": [1, 2]}], (), ()),
            (
                [{"function": "subtotal", "anchor": 9, "args": [1, 2], "name": "A"}],
                {3, 4},
                (),
            ),
            (
                [{"function": "subtotal", "anchor": 9, "args": [1, 2], "name": "B"}],
                {1, 2, 3, 4, 5, 8, -1},
                ({"function": "subtotal", "anchor": 9, "args": [1, 2], "name": "B"},),
            ),
            (
                [
                    {"function": "subtotal", "anchor": 9, "args": [1, 2], "name": "C"},
                    {"function": "subtotal", "anchor": 9, "args": [3, 4], "name": "Z"},
                    {"function": "subtotal", "anchor": 9, "args": [5, 6], "name": "D"},
                ],
                {1, 2, 5, 8, -1},
                (
                    {"function": "subtotal", "anchor": 9, "args": [1, 2], "name": "C"},
                    {"function": "subtotal", "anchor": 9, "args": [5, 6], "name": "D"},
                ),
            ),
        ),
    )
    def it_iterates_the_valid_subtotal_insertion_dicts_to_help(
        self, insertion_dicts, element_ids, expected_value, _element_ids_prop_
    ):
        _element_ids_prop_.return_value = element_ids
        subtotals = _Subtotals(insertion_dicts, None, None)

        subtotal_dicts = tuple(subtotals._iter_valid_subtotal_dicts())

        assert subtotal_dicts == expected_value

    def it_constructs_its_subtotal_objects_to_help(
        self, request, _iter_valid_subtotal_dicts_, valid_elements_, _Subtotal_
    ):
        subtotal_dicts_ = tuple({"subtotal-dict": idx} for idx in range(3))
        subtotal_objs_ = tuple(
            instance_mock(request, _Subtotal, name="subtotal-%d" % idx)
            for idx in range(3)
        )
        _iter_valid_subtotal_dicts_.return_value = iter(subtotal_dicts_)
        _Subtotal_.side_effect = iter(subtotal_objs_)
        subtotals = _Subtotals(None, valid_elements_, True)

        subtotal_objs = subtotals._subtotals

        assert _Subtotal_.call_args_list == [
            call(subtot_dict_, valid_elements_, True)
            for subtot_dict_ in subtotal_dicts_
        ]
        assert subtotal_objs == subtotal_objs_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _element_ids_prop_(self, request):
        return property_mock(request, _Subtotals, "_element_ids")

    @pytest.fixture
    def _iter_valid_subtotal_dicts_(self, request):
        return method_mock(request, _Subtotals, "_iter_valid_subtotal_dicts")

    @pytest.fixture
    def _Subtotal_(self, request):
        return class_mock(request, "cr.cube.dimension._Subtotal")

    @pytest.fixture
    def _subtotals_prop_(self, request):
        return property_mock(request, _Subtotals, "_subtotals")

    @pytest.fixture
    def valid_elements_(self, request):
        return instance_mock(request, _ValidElements)


class Describe_Subtotal(object):
    """Unit-test suite for `cr.cube.dimension._Subtotal` object."""

    @pytest.mark.parametrize(
        "subtotal_dict, addend_ids, expected_value",
        (
            ({"args": []}, (), []),
            ({"args": [1, 2]}, (1, 2), [0, 1]),
            ({"args": [0, 2]}, (1, 2), [0, 1]),
            ({"args": [3, 4]}, (3, 4), [2, 3]),
            ({"args": [4, 99]}, (4, 99), [3, 4]),
        ),
    )
    def it_knows_the_addend_idxs(
        self,
        addend_ids_,
        all_elements_,
        valid_elements_,
        subtotal_dict,
        addend_ids,
        expected_value,
    ):
        addend_ids_.return_value = addend_ids
        elements_ = (
            _Element({"id": 1}, None, None),
            _Element({"id": 2}, None, None),
            _Element({"id": 3}, None, None),
            _Element({"id": 4}, None, None),
            _Element({"id": 99}, None, None),
        )
        all_elements_.__iter__.return_value = iter(elements_)
        valid_elements = _ValidElements(all_elements_, None)
        subtotal = _Subtotal(subtotal_dict, valid_elements, None)

        addend_idxs = subtotal.addend_idxs

        np.testing.assert_array_almost_equal(addend_idxs, expected_value)

    @pytest.mark.parametrize(
        "subtotal_dict, element_ids, expected_value",
        (
            ({"anchor": 1}, {1, 2, 3}, 1),
            ({"anchor": 4}, {1, 2, 3}, "bottom"),
            ({"anchor": "Top"}, {1, 2, 3}, "top"),
            # For an undefined anchor default to "bottom"
            ({"anchor": None}, {1, 2, 3}, "bottom"),
        ),
    )
    def it_knows_the_insertion_anchor(
        self, subtotal_dict, element_ids, expected_value, valid_elements_
    ):
        valid_elements_.element_ids = element_ids
        subtotal = _Subtotal(subtotal_dict, valid_elements_, None)

        anchor = subtotal.anchor

        assert anchor == expected_value

    @pytest.mark.parametrize(
        "subtotal_dict, element_ids, expected_value",
        (
            ({}, {}, ()),
            ({"args": [1]}, {1}, (1,)),
            ({"args": [1, 2, 3]}, {1, 2, 3}, (1, 2, 3)),
            ({"args": [1, 2, 3]}, {1, 3}, (1, 3)),
            ({"args": [3, 2]}, {1, 2, 3}, (3, 2)),
            ({"args": []}, {1, 2, 3}, ()),
            ({"args": [1, 2, 3]}, {}, ()),
        ),
    )
    def it_provides_access_to_the_addend_element_ids(
        self, subtotal_dict, element_ids, expected_value, valid_elements_
    ):
        valid_elements_.element_ids = element_ids
        subtotal = _Subtotal(subtotal_dict, valid_elements_, None)

        addend_ids = subtotal.addend_ids

        assert addend_ids == expected_value

    @pytest.mark.parametrize(
        "subtotal_dict, expected_value",
        (({}, ""), ({"name": None}, ""), ({"name": ""}, ""), ({"name": "Joe"}, "Joe")),
    )
    def it_knows_the_subtotal_label(self, subtotal_dict, expected_value):
        subtotal = _Subtotal(subtotal_dict, None, None)

        label = subtotal.label

        assert label == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def addend_ids_(self, request):
        return property_mock(request, _Subtotal, "addend_ids")

    @pytest.fixture
    def all_elements_(self, request):
        return instance_mock(request, _AllElements)

    @pytest.fixture
    def valid_elements_(self, request):
        return instance_mock(request, _ValidElements)



import numpy as np

from cr.cube.cube import Cube

from ..fixtures import CR, SM  # ---mnemonic: SM = 'scale means'---


def test_ca_cat_x_items():
    slice_ = Cube(SM.CA_CAT_X_ITEMS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.50454821, 3.11233766, 3.35788192, 3.33271833]
    )
    assert slice_.scale_means_column is None
    assert slice_.scale_means_columns_margin is None


def test_ca_items_x_cat():
    slice_ = Cube(SM.CA_ITEMS_X_CAT).partitions[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.50454821, 3.11233766, 3.35788192, 3.33271833]
    )
    assert slice_.scale_means_rows_margin is None


def test_ca_itmes_x_cat_var_scale_means():
    # These 2 fixtures represent 1 dataset and its transpose version
    slice_ = Cube(SM.CA_ITEMS_X_CAT).partitions[0]
    slice2_ = Cube(SM.CA_CAT_X_ITEMS).partitions[0]

    # Testing that the scale means (row and col) are equal on the 2 diverse
    # datasets
    np.testing.assert_array_almost_equal(
        slice_.var_scale_means_column, slice2_.var_scale_means_row
    )

    np.testing.assert_array_almost_equal(
        slice2_.var_scale_means_row, [2.56410909, 5.17893869, 4.75445248, 4.81611278]
    )
    np.testing.assert_array_almost_equal(
        slice_.var_scale_means_column, [2.56410909, 5.17893869, 4.75445248, 4.81611278]
    )

    assert slice2_.var_scale_means_column is None
    assert slice_.var_scale_means_row is None


def test_ca_x_mr():
    slice_ = Cube(SM.CA_X_MR).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.29787234, 1.8, 1.48730964, np.nan]
    )
    assert slice_.scale_means_column is None
    assert slice_.scale_means_columns_margin is None
    assert slice_.scale_means_rows_margin == 1.504548211036992

    slice_ = Cube(SM.CA_X_MR).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [3.31746032, 3.10743802, 3.09976976, np.nan]
    )
    assert slice_.scale_means_column is None

    slice_ = Cube(SM.CA_X_MR).partitions[2]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [3.31205674, 3.23913043, 3.37745455, np.nan]
    )
    assert slice_.scale_means_column is None

    slice_ = Cube(SM.CA_X_MR).partitions[3]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [3.53676471, 3.34814815, 3.3147877, np.nan]
    )
    assert slice_.scale_means_column is None


def test_cat_x_ca_cat_x_items():
    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.34545455, 2.46938776, 2.7037037, 2.65454545]
    )
    assert slice_.scale_means_column is None
    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.41935484, 3.25663717, 3.48, 3.58536585]
    )
    assert slice_.scale_means_column is None
    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[2]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.49429038, 3.44905009, 3.59344262, 3.53630363]
    )
    assert slice_.scale_means_column is None
    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[3]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.43365696, 3.02816901, 3.37987013, 3.32107023]
    )
    assert slice_.scale_means_column is None
    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[4]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.22670025, 2.49473684, 2.79848866, 2.78987342]
    )
    assert slice_.scale_means_column is None
    slice_ = Cube(SM.CAT_X_CA_CAT_X_ITEMS).partitions[5]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.53061224, 3.68421053, 3.9862069, 4.03472222]
    )
    assert slice_.scale_means_column is None


def test_cat_x_cat():
    slice_ = Cube(SM.CAT_X_CAT).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.6009281, 2.3522267, 2.3197279, 3.3949192]
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [1.43636364, 2.45238095, 2.4730832, 2.68387097, 2.8375, 2.15540541],
    )

    # Test ScaleMeans marginal
    assert slice_.scale_means_columns_margin == 2.536319612590799
    assert slice_.scale_means_rows_margin == 2.6846246973365617


def test_cat_hs_x_cat_hs_var_scale_means():
    slice_ = Cube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS).partitions[0]

    assert slice_.var_scale_means_column is not None
    assert slice_.var_scale_means_row is not None

    np.testing.assert_almost_equal(
        slice_.var_scale_means_column,
        [0.88930748, 0.93655622, 1.36425875, 0.96388566, 3.55555556, 2.55601211],
    )
    np.testing.assert_almost_equal(
        slice_.var_scale_means_row,
        [
            0.51774691,
            0.51796281,
            0.99555556,
            0.84071826,
            1.12549449,
            1.19867769,
            2.4775,
        ],
    )


def test_cat_x_mr():
    slice_ = Cube(SM.CAT_X_MR).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.45070423, 2.54471545, 2.54263006, np.nan]
    )
    assert slice_.scale_means_column is None

    assert slice_.scale_means_columns_margin is None
    assert slice_.scale_means_rows_margin == 2.5323565323565322


def test_cat_x_cat_with_hs():
    # Test without H&S
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = Cube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85],
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [3.87368421, 2.51767677, 3.38429752, 3.66666667, 4.13235294],
    )

    # Test with H&S
    slice_ = Cube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85],
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [3.87368421, 2.51767677, 3.0851689, 3.38429752, 3.66666667, 4.13235294],
    )


def test_ca_x_mr_with_hs_and_pruning():
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.50818336, 2.56844883, 2.90251939, np.nan]
    )
    assert slice_.scale_means_column is None
    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.78385708, 2.69292009, 3.11594714, np.nan]
    )
    assert slice_.scale_means_column is None
    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[2]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [np.nan, np.nan, np.nan, np.nan]
    )
    assert slice_.scale_means_column is None

    transforms = {
        "rows_dimension": {"prune": True},
        "columns_dimension": {"prune": True},
    }
    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.50818336, 2.56844883, 2.90251939]
    )
    assert slice_.scale_means_column is None
    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [2.78385708, 2.69292009, 3.11594714]
    )
    assert slice_.scale_means_column is None
    slice_ = Cube(CR.CA_X_MR_HS, transforms=transforms).partitions[2]
    np.testing.assert_almost_equal(slice_.scale_means_row, [])
    assert slice_.scale_means_column is None


def test_cat_x_cat_pruning_and_hs():
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.57933884, 2.10618401, 2.30460074, np.nan, 2.34680135]
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [1.74213625, 1.97, 2.45356177, 2.11838791, np.nan, 2.0],
    )

    # Just H&S
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [1.57933884, 1.8308135, 2.10618401, 2.30460074, np.nan, 2.34680135],
    ),
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [1.74213625, 2.2364515, 1.97, 2.45356177, 2.11838791, np.nan, 2.0],
    )

    # Just pruning
    transforms = {
        "rows_dimension": {"prune": True},
        "columns_dimension": {"prune": True},
    }
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [1.57933884, 1.83081353, 2.10618401, 2.30460074, 2.34680135],
    )
    np.testing.assert_almost_equal(
        slice_.scale_means_column,
        [1.74213625, 2.2364515, 1.97, 2.45356177, 2.11838791, 2.0],
    )

    # Pruning and H&S
    transforms = {
        "rows_dimension": {"insertions": {}, "prune": True},
        "columns_dimension": {"insertions": {}, "prune": True},
    }
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.57933884, 2.106184, 2.3046007, 2.34680135]
    ),
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.74213625, 1.97, 2.45356177, 2.11838791, 2.0]
    )


def test_cat_x_cat_scale_means_margin():
    slice_ = Cube(SM.CAT_X_CAT_SM_MARGIN).partitions[0]
    assert slice_.scale_means_rows_margin == 2.6846246973365617
    assert slice_.scale_means_columns_margin == 2.536319612590799


def test_cat_x_ca_subvar_scale_means():
    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_SUBVARS_FIRST).partitions[0]
    np.testing.assert_almost_equal(
        slice_.var_scale_means_row, [0.2054321, 0.24, 0.22558594]
    )
    assert slice_.var_scale_means_column is None

    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_SUBVARS_FIRST).partitions[1]
    np.testing.assert_almost_equal(
        slice_.var_scale_means_row, [0.2283737, 0.21, 0.21606648]
    )
    assert slice_.scale_means_column is None


def test_cat_x_cat_pruning_and_hs_var_scale_means():
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.var_scale_means_row,
        [1.4459092, 2.14619102, 2.40430987, np.nan, 0.87972883],
    )
    np.testing.assert_almost_equal(
        slice_.var_scale_means_column,
        [0.72358198, 0.9991, 1.87633763, 0.4859843, np.nan, 0.66666667],
    )

    # Just H&S
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.var_scale_means_row,
        [1.4459092, 1.8494177, 2.14619102, 2.40430987, np.nan, 0.87972883],
    ),
    np.testing.assert_almost_equal(
        slice_.var_scale_means_column,
        [0.72358198, 1.08423566, 0.9991, 1.87633763, 0.4859843, np.nan, 0.66666667],
    )

    # Just pruning
    transforms = {
        "rows_dimension": {"prune": True},
        "columns_dimension": {"prune": True},
    }
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.var_scale_means_row,
        [1.4459092, 1.8494177, 2.14619102, 2.40430987, 0.87972883],
    )
    np.testing.assert_almost_equal(
        slice_.var_scale_means_column,
        [0.72358198, 1.08423566, 0.9991, 1.87633763, 0.4859843, 0.66666667],
    )

    # Pruning and H&S
    transforms = {
        "rows_dimension": {"insertions": {}, "prune": True},
        "columns_dimension": {"insertions": {}, "prune": True},
    }
    slice_ = Cube(CR.CAT_X_CAT_PRUNING_HS, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(
        slice_.var_scale_means_row, [1.4459092, 2.14619102, 2.40430987, 0.87972883]
    ),
    np.testing.assert_almost_equal(
        slice_.var_scale_means_column,
        [0.72358198, 0.9991, 1.87633763, 0.4859843, 0.66666667],
    )


def test_cat_nps_numval_x_cat_var_scale_means():
    slice_ = Cube(SM.CAT_NPS_NUMVAL_X_CAT).partitions[0]
    np.testing.assert_almost_equal(
        slice_.var_scale_means_row,
        [1905.11600238, 2111.67820069, 1655.65636907, 981.86821176],
    )
    assert slice_.var_scale_means_column is None


def test_cat_single_element_x_cat():
    slice_ = Cube(SM.CAT_SINGLE_ELEMENT_X_CAT).partitions[0]
    np.testing.assert_equal(slice_.scale_means_row, [np.nan, np.nan, np.nan, np.nan])
    np.testing.assert_equal(slice_.scale_means_column, [np.nan])


def test_means_univariate_cat():
    strand = Cube(CR.ECON_BLAME_WITH_HS).partitions[0]
    np.testing.assert_almost_equal(strand.scale_mean, [2.1735205616850553])


def test_means_bivariate_cat():
    slice_ = Cube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row,
        [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85],
    )


def test_means_cat_x_mr():
    slice_ = Cube(CR.FRUIT_X_PETS).partitions[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.7, 1.6470588, 1.6842105])
    assert slice_.scale_means_column is None


def test_means_mr_x_cat():
    slice_ = Cube(CR.PETS_X_FRUIT).partitions[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.7, 1.6470588, 1.6842105]
    )


def test_means_cat_array_cat_dim_first():
    slice_ = Cube(CR.PETS_ARRAY_CAT_FIRST).partitions[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.44333002, 1.48049069, 1.57881177]
    )


def test_means_cat_array_subvar_dim_first():
    slice_ = Cube(CR.PETS_ARRAY_SUBVAR_FIRST).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.44333002, 1.48049069, 1.57881177]
    )
    assert slice_.scale_means_column is None


def test_means_cat_x_cat_arr_fruit_first():
    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY).partitions[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.48, 1.4285714, 1.5217391]
    )
    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY).partitions[1]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [1.40740741, 1.53846154, 1.55319149]
    )


def test_means_cat_x_cat_arr_subvars_first():
    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_SUBVARS_FIRST).partitions[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.71111111, 1.6, 1.65625])
    assert slice_.scale_means_column is None

    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_SUBVARS_FIRST).partitions[1]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.64705882, 1.7, 1.68421053]
    )
    assert slice_.scale_means_column is None


def test_means_cat_x_cat_arr_pets_first():
    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST).partitions[0]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.48, 1.40740741])
    np.testing.assert_almost_equal(slice_.scale_means_column, [1.71111111, 1.64705882])

    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST).partitions[1]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.42857143, 1.53846154])
    np.testing.assert_almost_equal(slice_.scale_means_column, [1.6, 1.7])

    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST).partitions[2]
    np.testing.assert_almost_equal(slice_.scale_means_row, [1.52173913, 1.55319149])
    np.testing.assert_almost_equal(slice_.scale_means_column, [1.65625, 1.68421053])


def test_means_with_null_values():
    slice_ = Cube(CR.SCALE_WITH_NULL_VALUES).partitions[0]
    np.testing.assert_almost_equal(
        slice_.scale_means_row, [1.2060688, 1.0669344, 1.023199]
    )
    assert slice_.scale_means_column is None


def test_mean_univariate_cat_var_scale_mean():
    # Test nonmissing with no null numeric values
    strand_ = Cube(SM.UNIVARIATE_CAT).partitions[0]
    is_a_number_mask = ~np.isnan(strand_._numeric_values)

    np.testing.assert_almost_equal(strand_.var_scale_mean, 5.4322590719809645)
    np.testing.assert_array_equal(
        strand_._counts_as_array[is_a_number_mask], [54, 124, 610, 306, 396, 150]
    )

    # Test nonmissing with null numeric value
    strand_ = Cube(SM.UNIVARIATE_CAT_WITH_NULL_NUMERIC_VALUE).partitions[0]
    is_a_number_mask = ~np.isnan(strand_._numeric_values)

    np.testing.assert_almost_equal(strand_.var_scale_mean, 5.517066895232401)
    np.testing.assert_array_equal(
        strand_._counts_as_array[is_a_number_mask], [124, 610, 306, 396, 150]
    )

    # Test with all null numeric value
    strand_ = Cube(SM.UNIVARIATE_CAT_WITH_ALL_NULL_NUMERIC_VALUE).partitions[0]
    is_a_number_mask = ~np.isnan(strand_._numeric_values)

    assert strand_.var_scale_mean is None
    np.testing.assert_array_equal(
        is_a_number_mask, np.array([False] * len(strand_._numeric_values))
    )
    np.testing.assert_array_equal(strand_._counts_as_array[is_a_number_mask], [])


def test_mr_x_cat():
    slice_ = Cube(SM.MR_X_CAT).partitions[0]
    assert slice_.scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.scale_means_column, [2.45070423, 2.54471545, 2.54263006, np.nan]
    )

    assert slice_.scale_means_columns_margin == 2.5323565323565322
    assert slice_.scale_means_rows_margin is None


def test_var_scale_means_for_fruit_x_pets_array():
    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY).partitions[0]

    assert slice_.var_scale_means_row is None
    np.testing.assert_almost_equal(
        slice_.var_scale_means_column, np.array([0.2496, 0.24489796, 0.24952741])
    )

    slice_ = Cube(CR.FRUIT_X_PETS_ARRAY).partitions[1]

    np.testing.assert_almost_equal(
        slice_.var_scale_means_column, np.array([0.24142661, 0.24852071, 0.24717067])
    )


def test_univariate_cat():
    strand = Cube(SM.UNIVARIATE_CAT).partitions[0]
    np.testing.assert_almost_equal(strand.scale_mean, [2.6865854])


def test_univariate_cat_with_hiding():
    strand_ = Cube(SM.BOLSHEVIK_HAIR).partitions[0]
    np.testing.assert_almost_equal(strand_.scale_mean, [1.504548211])

    # Appling hiding transforms
    transforms = {
        "rows_dimension": {"elements": {"5": {"hide": True}, "4": {"hide": True}}}
    }
    strand_with_hiding_ = Cube(SM.BOLSHEVIK_HAIR, transforms=transforms).partitions[0]
    np.testing.assert_almost_equal(strand_.scale_mean, strand_with_hiding_.scale_mean)


def test_univariate_with_hs():
    # Test without H&S
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    strand = Cube(CR.ECON_BLAME_WITH_HS, transforms).partitions[0]
    np.testing.assert_almost_equal(strand.scale_mean, [2.17352056])

    # Test with H&S
    strand = Cube(CR.ECON_BLAME_WITH_HS).partitions[0]
    np.testing.assert_almost_equal(strand.scale_mean, [2.17352056])


def test_univariate_with_hs_var_scale_means_row():
    # Test without H&S
    transforms = {
        "columns_dimension": {"insertions": {}},
        "rows_dimension": {"insertions": {}},
    }
    strand = Cube(CR.ECON_BLAME_WITH_HS, transforms).partitions[0]
    np.testing.assert_almost_equal(strand.var_scale_mean, [1.1363901131679894])

    # Test with H&S
    strand = Cube(CR.ECON_BLAME_WITH_HS).partitions[0]
    np.testing.assert_almost_equal(strand.var_scale_mean, [1.1363901131679894])



# encoding: utf-8

"""Integration test suite for cr.cube.cube_slice module."""

import numpy as np

from cr.cube.crunch_cube import CrunchCube

from ...fixtures import CR  # ---mnemonic: CR = 'cube-response'---


class DescribeIntegratedCubeSlice(object):
    def it_provides_a_console_friendly_repr_for_a_slice(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        slice_ = cube.slices[0]

        repr_ = repr(slice_)

        assert repr_ == (
            "CubeSlice(name='v4', dim_types='CAT x CAT', dims='v4 x v7')\n"
            "      C    E\n"
            "--  ---  ---\n"
            "B     5    2\n"
            "C     5    3"
        )

    def it_knows_zscore_for_MR_X_MR_X_CAT(self):
        cube = CrunchCube(CR.MR_X_MR_X_CAT)
        slice_ = cube.slices[0]

        zscore = slice_.zscore()

        np.testing.assert_array_almost_equal(
            zscore,
            np.array(
                [
                    [-2.0890161, 2.0890161],
                    [-0.31167807, 0.31167807],
                    [-0.4574511, 0.4574511],
                    [0.08713899, -0.08713899],
                    [-1.88534457, 1.88534457],
                    [-1.18201963, 1.18201963],
                    [-0.14570511, 0.14570511],
                    [-0.79331024, 0.79331024],
                    [-0.10870154, 0.10870154],
                    [0.39665764, -0.39665764],
                    [-0.76214626, 0.76214626],
                    [-1.59164242, 1.59164242],
                    [np.nan, np.nan],
                    [-1.97343777, 1.97343777],
                    [-0.63278752, 0.63278752],
                    [-0.4446455, 0.4446455],
                    [-0.10084772, 0.10084772],
                    [-0.42861964, 0.42861964],
                ]
            ),
        )

    def test_mr_x_mr_single_cat_pruning(self):
        cc = CrunchCube(CR.MR_X_MR_SINGLE_CAT)
        cube_slice = cc.slices[0]
        expected = np.ma.masked_array([2, 1, 0], mask=[False, False, True])
        np.testing.assert_array_equal(cube_slice.as_array(prune=True), expected)

    def test_mr_single_cat_x_mr_margin_with_pruning(self):
        cc = CrunchCube(CR.MR_SINGLE_CAT_X_MR)
        cube_slice = cc.slices[0]
        expected = np.ma.masked_array([2, 2, 2], mask=[False, False, True])
        np.testing.assert_array_equal(cube_slice.margin(prune=True), expected)


class DescribeCubeSliceAPI(object):
    def it_calculates_correct_index_tables_for_single_elements(self):
        cs = CrunchCube(CR.MR_X_CAT_BOTH_SINGLE_ELEMENT).slices[0]

        # Check for column direction (as in the exporter), backed by R
        expected = np.array([[0.0, 110.39006714]])
        index_table = cs.index_table(axis=0)
        np.testing.assert_array_almost_equal(index_table, expected)

        # Check for row direction, backed by R
        expected = np.array([[0.0, 116.472612]])
        index_table = cs.index_table(axis=1)
        np.testing.assert_array_almost_equal(index_table, expected)


def test_labels_with_hs_and_pruning():
    cs = CrunchCube(CR.CAT_X_CAT_PRUNING_HS).slices[0]

    # Withouut pruning or H&S
    expected = [
        [
            u"Married",
            u"Separated",
            u"Divorced",
            u"Widowed",
            u"Single",
            u"Domestic partnership",
        ],
        [
            u"President Obama",
            u"Republicans in Congress",
            u"Both",
            u"Neither",
            u"Not sure",
        ],
    ]
    actual = cs.labels()
    assert actual == expected

    # Apply pruning
    expected = [
        [u"Married", u"Separated", u"Divorced", u"Widowed", u"Domestic partnership"],
        [u"President Obama", u"Republicans in Congress", u"Both", u"Not sure"],
    ]
    actual = cs.labels(prune=True)
    assert actual == expected

    # Apply H&S
    expected = [
        [
            u"Married",
            u"left alone",
            u"Separated",
            u"Divorced",
            u"Widowed",
            u"Single",
            u"Domestic partnership",
        ],
        [
            u"President Obama",
            u"Obama + Republicans",
            u"Republicans in Congress",
            u"Both",
            u"Neither",
            u"Not sure",
        ],
    ]
    actual = cs.labels(hs_dims=[0, 1])
    assert actual == expected

    # Apply H&S and pruning
    expected = [
        [
            u"Married",
            u"left alone",
            u"Separated",
            u"Divorced",
            u"Widowed",
            u"Domestic partnership",
        ],
        [
            u"President Obama",
            u"Obama + Republicans",
            u"Republicans in Congress",
            u"Both",
            u"Not sure",
        ],
    ]
    actual = cs.labels(prune=True, hs_dims=[0, 1])
    assert actual == expected
